import unittest
from rdkit import Chem
from fragmenstein import Monster
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem
from fragmenstein.monster.mcs_mapping import SpecialCompareAtoms
# ======================================================================================================================

class Mappings(unittest.TestCase):
    """
    These test monster.mcs_mapping
    """
    toluene = Chem.MolFromSmiles('Cc1ccccc1')
    toluene.SetProp('_Name', 'toluene')

    benzyl = Chem.MolFromSmiles('*c1ccccc1')
    benzyl.SetProp('_Name', 'benzyl')

    methylpyrylium = Chem.MolFromSmiles('Cc1c[o+]ccc1')
    methylpyrylium.SetProp('_Name', 'methylpyrylium')
    # Monster.draw_nicely(None, methylpyrylium)

    methylpyridine = Chem.MolFromSmiles('Cc1ncccc1')
    methylpyridine.SetProp('_Name', 'methylpyridine')
    # Monster.draw_nicely(None, methylpyridine)


    def test_not_to_dummy(self):
        from rdkit.Chem import rdFMCS
        params = rdFMCS.MCSParameters()  # noqa
        params.BondTyper = rdFMCS.BondCompare.CompareAny
        params.AtomTyper = SpecialCompareAtoms()
        compare = [self.benzyl, self.toluene]
        # hit -> followup
        res: rdFMCS.MCSResult = rdFMCS.FindMCS(compare, params)
        print(res.smartsString)
        self.assertEqual(res.numAtoms, 6)  # there are 7 atoms, but only 6 are mapped as the dummy is excluded

    def test_user_map(self):
        """
        Map the azo with the oxo in pyridine and pyrylium.
        Using the methyl versions, the methyl does not map.
        """
        from rdkit.Chem import rdFMCS
        custom_map = {'methylpyridine': {2: 3}}
        params = rdFMCS.MCSParameters()
        params.BondTyper = rdFMCS.BondCompare.CompareAny
        params.AtomTyper = SpecialCompareAtoms(custom_map=custom_map)
        compare = [self.methylpyridine, self.methylpyrylium]
        # hit -> followup
        res: rdFMCS.MCSResult = rdFMCS.FindMCS(compare, params)
        self.assertEqual(res.numAtoms, 6, 'Mapping  the azo with the oxo means no methyl res:' +
                                         f' 6 matched, not {res.numAtoms}')
        full_maps = params.AtomTyper.get_valid_matches(params.AtomCompareParameters,
                                                      Chem.MolFromSmarts(res.smartsString),
                                                      self.methylpyridine, self.methylpyrylium
                                                      )
        # this is wrong: you can map the methyls to either side (ortho) of the heteroatom
        # self.assertEqual(len(full_maps), 1, f'There is only one way to map it not {len(full_maps)} ({full_maps})')
        for full_map in full_maps:
            for h, f in custom_map['methylpyridine'].items():
                self.assertEqual(dict(full_map)[h], f, f'hit atom {h} => {f} not {dict(full_map)[h]}')

    def test_user_negmap(self):
        """
        Ban index 1 on the azo=oxo mapping
        """
        from rdkit.Chem import rdFMCS
        custom_map = {'methylpyridine': {2: 3, 1: -2}}
        params = rdFMCS.MCSParameters()
        params.BondTyper = rdFMCS.BondCompare.CompareAny
        params.AtomTyper = SpecialCompareAtoms(custom_map=custom_map)
        compare = [self.methylpyridine, self.methylpyrylium]
        # hit -> followup
        res: rdFMCS.MCSResult = rdFMCS.FindMCS(compare, params)
        self.assertEqual(res.numAtoms, 5, 'ought to be azo=oxo and no index 1')
        full_maps = params.AtomTyper.get_valid_matches(params.AtomCompareParameters,
                                                       Chem.MolFromSmarts(res.smartsString),
                                                       self.methylpyridine, self.methylpyrylium
                                                       )
        # this is wrong: you can map the methyls to either side (ortho) of the heteroatom
        # self.assertEqual(len(full_maps), 1, f'There is only one way to map it not {len(full_maps)} ({full_maps})')
        for full_map in full_maps:
            self.assertNotIn(1, dict(full_map).keys())

    def test_flipper(self):
        from fragmenstein.monster.mcs_mapping import flip_mapping
        # Sequence[Tuple[int, int]]
        flipped = flip_mapping([(1, 2)])
        self.assertIsInstance(flipped, (list, tuple))


    def test_user_negmap2(self):
        """
        Assign followup index 1 to another molecule ('nullium' as its nothing) in the azo=oxo mapping
        thus preventing its mapping to methylpyridine
        """
        from rdkit.Chem import rdFMCS
        custom_map = {'methylpyridine': {2: 3}, 'nullium': {1: 1}}
        params = rdFMCS.MCSParameters()
        params.BondTyper = rdFMCS.BondCompare.CompareAny
        params.AtomTyper = SpecialCompareAtoms(custom_map=custom_map)
        compare = [self.methylpyridine, self.methylpyrylium]
        # hit -> followup
        res: rdFMCS.MCSResult = rdFMCS.FindMCS(compare, params)
        self.assertEqual(res.numAtoms, 5, 'ought to be azo=oxo and no index 1')
        full_maps = params.AtomTyper.get_valid_matches(params.AtomCompareParameters,
                                                       Chem.MolFromSmarts(res.smartsString),
                                                       self.methylpyridine, self.methylpyrylium
                                                       )
        # this is wrong: you can map the methyls to either side (ortho) of the heteroatom
        # self.assertEqual(len(full_maps), 1, f'There is only one way to map it not {len(full_maps)} ({full_maps})')
        for full_map in full_maps:
            self.assertNotIn(1, dict(full_map).values())

    def test_issue42(self):
        RDLogger.DisableLog('rdApp.warning')  # shut up about the hydrogens already
        # hydrogenated indole w/ a methyl on the r6
        hit = Chem.MolFromSmiles('N1CCC2C1CC(C)CC2')
        AllChem.EmbedMolecule(hit)
        hit.SetProp('_Name', 'foo')
        monstah = Monster([hit])
        # hydrogenated benzothiazole w/ no methyl
        monstah.place_smiles('N1CSC2C1CCCC2')
        self.assertEqual('foo.0+foo.1+foo.2+foo.3+foo.4+foo.5+foo.6+foo.8+foo.9',
                         '+'.join([o[0] if o else 'NA' for o in monstah.origin_from_mol()]))



if __name__ == '__main__':
    unittest.main()



import numpy as np
from typing import Optional
from rdkit import Geometry, Chem

class WaltonPolygon:

    @staticmethod
    def create_polygon(n=6, radius=1.5,
                       x0=0., y0=0., z0=0.,
                       star=False,  # technical polygram?
                       bond_type: Chem.BondType = Chem.BondType.SINGLE,
                       starting_mol: Optional[Chem.Mol] = None,
                       name='polygon'):
        """
        Creates a xy flat polygon or star of radius ``radius`` with ``n`` sides.
        """
        if starting_mol:
            mol = Chem.RWMol(starting_mol)
            conf = starting_mol.GetConformer()
        else:
            mol = Chem.RWMol()
            conf = Chem.Conformer()
        center = -1  # just to avoid #noqas
        if star:
            center: int = mol.AddAtom(Chem.Atom(6))
            conf.SetAtomPosition(center, Geometry.Point3D(x0, y0, z0))
        for i in range(n):
            j = mol.AddAtom(Chem.Atom(6))
            if star:
                mol.AddBond(center, j, bond_type)
            elif j == 0:  # no bond yet
                pass
            else:
                mol.AddBond(j - 1, j, bond_type)
            conf.SetAtomPosition(j, Geometry.Point3D(x0 + radius * np.cos(i * 2 * np.pi / n),
                                                     y0 + radius * np.sin(i * 2 * np.pi / n),
                                                     z0)
                                 )
        if not star:
            # bond last atom to first:
            mol.AddBond(j - n + 1, j, bond_type)
        mol.AddConformer(conf)
        mol = mol.GetMol()
        mol.SetProp('_Name', name)
        return mol



import os
from io import StringIO, BytesIO
from rdkit import Chem
from rdkit.Chem import Descriptors
from typing import List, Optional
from types import ModuleType
import importlib.resources as pkg_resources
import random


class BaseMolDataset:
    """
    Helper functions for loading the data.
    I had considered making it a "module factory" but gets quickly unintelligible.
    """

    dataset_package: ModuleType = '__main__'

    def __init__(self):
        raise NotImplementedError('Virtual method')

    def raise_not_implemented(self):
        raise NotImplementedError('Method not implemented for this dataset')

    @classmethod
    def file2mol_func(cls, filename: str) -> str:
        return os.path.splitext(filename)[0].strip()

    @classmethod
    def mol2file_func(cls, hit_name: str) -> str:
        return f'{hit_name.strip()}.mol'

    @classmethod
    def get_molblock(cls, hit_name: Optional[str] = None) -> str:
        """
        returns the mol block for the given hit. Chosen randomly if unspecified.
        See ``get_mpro_hit_list()`` for options.
        """
        if hit_name is None:
            hit_name = random.choice(cls.get_hit_list())
        return cls.get_text(cls.mol2file_func(hit_name))
    
    @classmethod
    def get_hit_list(cls) -> List[str]:
        return [cls.file2mol_func(fn) for fn in pkg_resources.contents(cls.dataset_package) if '.mol' in fn]
    
    @classmethod
    def get_text(cls, filename: str) -> str:
        return pkg_resources.read_text(cls.dataset_package, filename)

    @classmethod
    def get_bytes(cls, filename: str) -> bytes:
        return pkg_resources.read_binary(cls.dataset_package, filename)
    
    @classmethod
    def exists(cls, filename: str) -> bool:
        return pkg_resources.is_resource(cls.dataset_package, filename)

    @classmethod
    def get_mol(cls, hit_name: Optional[str] = None) -> Chem.Mol:
        """
        returns the Chem.Mol instance for the given hit. Chosen randomly if unspecified.
        See ``get_mpro_hit_list()`` for options.
        """
        mol = Chem.MolFromMolBlock(cls.get_molblock(hit_name))
        mol.SetProp('_Name', hit_name)
        return mol

    @classmethod
    def get_sdf_mols(cls, filename: str) -> List[Chem.Mol]:
        """
        Get the Chem.Mols in the specified sdf file.

        :param filename:
        :return:
        """
        bytes_io = BytesIO(cls.get_bytes(filename))
        with Chem.ForwardSDMolSupplier(bytes_io) as reader:
            return list(reader)

    @classmethod
    def get_filtered_mol(cls, mw_cutoff: float = float('inf')) -> Chem.Mol:
        """
        Return a random Chem.Mol from the MPro dataset,
        with one of the following filtering criteria:

        * ``mw_cutoff``: molWt cutoff
        """
        # For now only MolWt
        choices = cls.get_hit_list()
        random.shuffle(choices)
        for hit_name in choices:
            mol = cls.get_mol(hit_name)
            if mw_cutoff >= Descriptors.MolWt(mol):
                return mol
        else:
            raise ValueError('No hit matches the specified values')

    @classmethod
    def get_n_filtered_mols(cls, amount: int, **cutoffs) -> List[Chem.Mol]:
        """Get ``amount`` of the mols (Chem.Mol) randomly
         that match a cutoff criterion.
        As listed in ``get_filtered_mol``"""
        mols = []
        mol_names = set()
        assert amount > 0, 'A zero amount does nothing.'
        for i in range(len(cls.get_hit_list())):
            mol = cls.get_filtered_mol(**cutoffs)
            mol_name = mol.GetProp('_Name')
            if mol_name in mol_names:
                continue
            mol_names.add(mol_name)
            mols.append(mol)
            if len(mols) == int(amount):
                return mols
        else:
            raise ValueError(f'There are only {len(mols)} molecules ({mol_names}) that match the criteria {cutoffs}')

    @classmethod
    def get_template(cls) -> str:
        return cls.get_text('template.pdb')

# not for the classes to use
def extend_doc(cls):
    """
    This adds usage notes dynamically to the modules.
    The spacing will be off.
    """
    extra = f"""
    The module {cls.dataset_package.__name__} contains the data, but
    not functions to retrieve them without using ``pkg_resources``.
    These are added by the class {cls.__name__} as class methods.
    Not all of the methods are implemented for all datasets.
    """
    cls.dataset_package.__doc__ += extra



from __future__ import annotations
from enum import Enum
from rdkit import Chem
from typing import List

class BondProvenance(Enum):
    """
    Where does the bond come from. This is used to keep names consistent...
    For now original is used in places. The others are interchangeable TBH.
    """
    ORIGINAL = 1 # present in original mols
    MAIN_NOVEL = 2 # joined (zero-atom linking). Closest
    OTHER_NOVEL = 3 # joined (zero-atom linking). Others
    LINKER = 4
    UNASSIGNED = 5 # likely an error

    @classmethod
    def set_all_bonds(cls, mol: Chem.Mol, provenance_name:str) -> None:
        """
        Sets the provenance of all bonds in mol to a category, which is a string from the provenance

        :param mol:
        :param provenance_name: A string original | main_novel " other_novel | linker
        :return:
        """
        for bond in mol.GetBonds():
            cls.set_bond(bond, provenance_name)


    @classmethod
    def get_bond(cls, bond: Chem.Bond) -> BondProvenance:
        if not bond.HasProp('_Provenance'):
            return cls.UNASSIGNED
        else:
            p = bond.GetIntProp('_Provenance')
            return cls(p)

    @classmethod
    def get_bonds(cls, bonds: List[Chem.Bond]) -> List[BondProvenance]:
        return [cls.get_bond(bond) for bond in bonds]

    @classmethod
    def set_bond(cls, bond: Chem.Bond, provenance_name: str) -> None:
        bond.SetIntProp('_Provenance', cls[provenance_name.upper()].value)

    @classmethod
    def set_bonds(cls, bonds: List[Chem.Bond], provenance_name: str) -> None:
        for bond in bonds:
            cls.set_bond(bond, provenance_name)

    @classmethod
    def has_bond(cls, bond: Chem.Bond) -> bool:
        return bond.HasProp('_Provenance')

    @classmethod
    def copy_bond(cls, donor: Chem.Bond, acceptor: Chem.Bond) -> None:
        p = cls.get_bond(donor)
        acceptor.SetIntProp('_Provenance', p.value)






import unittest
import subprocess
import os

import pandas as pd

from fragmenstein import Victor
from fragmenstein.cli import FragmensteinParser
from fragmenstein import Laboratory
import fragmenstein.demo.test_mols as mol_folder
from fragmenstein.demo import Mac1
from pathlib import Path
from rdkit import Chem


# from tempfile import NamedTemporaryFile

class CliTests(unittest.TestCase):

    def command(self, command):
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        return out, err, process.returncode

    def test_monster_combine(self):
        parser = FragmensteinParser()
        return parser(['monster', 'combine',
                       '-i', str(Path(mol_folder.__file__).parent / 'mac-x0138.mol'),
                       str(Path(mol_folder.__file__).parent / 'mac-x0398.mol')
                       ])

    def test_monster_place(self):
        parser = FragmensteinParser()
        return parser(['monster', 'place',
                       '-i', str(Path(mol_folder.__file__).parent / 'mac-x0138.mol'),
                       str(Path(mol_folder.__file__).parent / 'mac-x0398.mol'),
                       '-s', 'COCCON'
                       ])

    def test_pipeline(self):
        # by default the results are written to the current working directory, yet we dont care here
        pdb_filename = (Path(Victor.work_path) / 'test.pdb').absolute().as_posix()
        sd_filename = (Path(Victor.work_path) / 'test.sdf').absolute().as_posix()
        os.chdir(Victor.work_path)
        print(  os.getcwd()  )
        # create the input files
        pdb_block = Mac1.get_template()
        hits = [Mac1.get_mol(f'diamond-{name}') for name in ['x0282_A', 'x0104_A', 'x0722_A']]  # , 'x0591_A', 'x0091_B'
        with Chem.SDWriter(sd_filename) as sdfh:
            for hit in hits:
                sdfh.write(hit)
        with open(pdb_filename, 'w') as pdbfh:
            pdbfh.write(pdb_block)
        # run the pipeline
        parser = FragmensteinParser()
        parser(['pipeline',
                '--input', sd_filename,
                '--template', pdb_filename,
                '--suffix', 'test',
                '--sw_length', '5',
                '--sw_databases', 'REAL-Database-22Q1.smi.anon'
                ])

    def zest_temp(self):
        """
        This was broken...
        """
        df_filename = (Path(Victor.work_path) / 'fragmenstein_placedtest.pkl.gz').absolute().as_posix()
        all_placements = pd.read_pickle(df_filename)
        Laboratory.export_sdf(df=all_placements)


"""
NGL may not be installed: cf https://github.com/matteoferla/Fragmenstein/issues/37
as it's an optional: NGLview breaks Colaboratory in Jan 2024.

Both Victor, Monster and Walton use this optionally via the method ``.to_nglview``
which is inherited from:
_MonsterUtilCompare.to_nglview
_VictorShow.to_nglview
WaltonArt.to_nglview
"""

from rdkit import Chem
import nglview as nv
from nglview.component import ComponentViewer
from io import StringIO
import warnings

class MolNGLWidget(nv.NGLWidget):
    """
    Adds:

    a method ``add_mol`` that simply adds an rdkit molecule to the viewer.

    .. code-block:: python

        from rdkit import Chem
        mol = Chem.MolFromSmiles('CC')
        view = MolNGLWidget()
        view.add_mol(mol, colorValue='pink')

    A method ``add_neighbors`` that adds a representation of the neighbours of a given selection.

    .. code-block:: python

        view.add_neighbors('[ATP]', comp_id=0, radius=5, style='hyperball', color='gainsboro')

    """

    def add_mol(self, mol: Chem.Mol, carbon_color: str = '', colorValue:str='') -> ComponentViewer:
        """
        Add a rdkit.Chem to an NGLWidget.
        This function is used by Walton and Monster

        :param mol: rdkit.Chem.Mol
        :param colorValue: if blank, the color in the property _color is used.
        :return:
        """
        if carbon_color:
            # this is to standardise the color names with py3Dmol
            colorValue = carbon_color
        if not mol:
            raise ValueError('Provided mol is None: ' + \
                             'if user manual tinkering happened, please run monster.fix_hits')
        fh = StringIO(Chem.MolToPDBBlock(mol))  # I want atom names if present
        comp: ComponentViewer = self.add_component(fh,  # noqa it's there.
                                                   name=mol.GetProp('_Name'),
                                                   ext='pdb'
                                                   )
        comp.remove_ball_and_stick()
        if not colorValue and mol.HasProp('color'):
            colorValue = mol.GetProp('color')
        if colorValue:
            comp.add_representation('ball+stick', colorValue=colorValue, multipleBond=True)
        else:
            comp.add_representation('ball+stick', multipleBond=True)
        # _color business stems from Walton.
        return comp

    def remove_all_components(self):
        """
        This removes the components and all traces of their existance in the widget

        :return:
        """
        self._js(f"""this.stage.removeAllComponents()""")
        i = 0
        while hasattr(self, f'component_{i}'):
            delattr(self, f'component_{i}')
            i += 1
        self._ngl_component_ids = []

    def add_neighbors(self, selection: str, comp_id: int = 0, radius: float = 5, style: str = 'hyperball',
                      color: str = 'gainsboro'):
        """
        Given a compounent id as an integer interpreted by NGL JS (not 'component_0'), show the neighbours.

        :param selection: NGL style selection string, e.g. ``[ATP] or 1-30:A``
        :param comp_id: integer not string
        :param radius: float of atom to atom distance max
        :param style: see NGL manual for the representation styles. Default is ``hyperball``
        :param color: hashtag prefixed hex color string or CSS color name
        :return:
        """
        self._js(f"""const comp = this.stage.compList[{comp_id}]
                     const target_sele = new NGL.Selection('{selection}');
                     const radius = {radius};
                     const neigh_atoms = comp.structure.getAtomSetWithinSelection( target_sele, radius );
                     const resi_atoms = comp.structure.getAtomSetWithinGroup( neigh_atoms );
                     comp.addRepresentation( "{style}", {{sele: resi_atoms.toSeleString(),
                     									  colorValue: "{color}",
                                                          multipleBond: true
                                                                       }});
                     comp.addRepresentation( "contact", {{sele: resi_atoms.toSeleString()}});
                """)

    def add_selection_signal(self, molname_id: str, atom_id: str):
        """
        Add a signal to the viewer than fills the elements #molname_id and #atom_id
        when user clicks on an atom —not bond.


        .. code-block:: python

            from IPython.display import display, HTML
            # in real code the html entities &gt;/&lt; would be actual greater-than and lesser-than signs
            display(HTML(f'&gt;div id="#{molname_id}"&lt;&gt;/div>&gt;div id="#{atom_id}"&lt;&gt;/div&lt;'))
            view.add_selection_signal(molname_id, atom_id)

        :param molname_id:
        :param atom_id:
        :return:
        """
        self._js(f'''this.stage.signals.clicked.add(pickingProxy => {{
                if (pickingProxy && (pickingProxy.atom || pickingProxy.bond )){{
                    const atom = pickingProxy.atom || pickingProxy.closestBondAtom;
                    const component = pickingProxy.component;
                    console.log(atom);
                    console.log(component);
                    document.getElementById('{molname_id}').innerText = `${{component.name}}`;
                    document.getElementById('{atom_id}').innerText = `${{atom.index}}`;
                    }}
            }});
        ''')


import functools
from typing import (Union, Iterator, Sequence, List)

import pandas as pd
import pebble
from rdkit import Chem

from ..igor import pyrosetta  # this may be pyrosetta or a mock for Sphinx in RTD

from typing import Dict, TypedDict, NotRequired  # noqa monkeypatched by .legacy. Absent in <Py3.8

from ._base import LabBench, binarize, unbinarize


class MolPlacementInput(TypedDict):
    """
    The place command accepts a list of dictionary or a DataFrame,
    with three keys.

    The problem is that hits needs to be pickled properly without losing Properties.
    Hence the conversion to BinPlacementInput

    """
    smiles: str
    name: str
    hits: Sequence[Chem.Mol]
    custom_map: NotRequired[Dict[str, Dict[int, int]]]


class BinPlacementInput(TypedDict):
    """
    Like MolPlacementInput but with binary hits

    """
    smiles: str
    name: str
    binary_hits: Sequence[bytes]
    custom_map: NotRequired[Dict[str, Dict[int, int]]]

class LabPlace(LabBench):

    def place_subprocess(self, inputs: BinPlacementInput):
        """
        This is the placement subprocess. The combination subprocess is ``place_subprocess``.
        They are very similar...
        """
        name: str = inputs['name']
        smiles: str = inputs['smiles']
        if self.Victor.uses_pyrosetta:
            pyrosetta.distributed.maybe_init(extra_options=self.init_options)
        try:
            binary_hits = inputs['binary_hits']
            hits: List[Chem.Mol] = [hit for hit in map(unbinarize, binary_hits) if hit]
            assert len(hits) > 0, 'No valid hits!'
            # `self.Victor` is likely `Victor` but the user may have switched for a subclass, cf. `VictorMock`...
            self.journal.debug(f'Using {self.Victor.__name__}')
            victor = self.Victor(hits=hits,
                            pdb_block=self.pdbblock,
                            ligand_resn='LIG',
                            ligand_resi=self.ligand_resi,
                            covalent_resi=self.covalent_resi,
                            **self.settings
                            )
            victor.place(smiles, long_name=name)
            result: dict = {**dict(inputs), **victor.summarize()}
            result['unmin_binary'] = binarize(victor.monster.positioned_mol)
            result['min_binary'] = binarize(victor.minimized_mol)
            result['hit_binaries'] = [binarize(h) for h in victor.hits]
            if self.run_plip and victor.minimized_pdbblock:
                result.update(victor.get_plip_interactions())
            return result
        except KeyboardInterrupt as err:
            raise err
        except Exception as error:
            error_msg = f'{error.__class__.__name__} {error}'
            self.Victor.journal.critical(f'*** {error_msg} for {name}')
            return dict(error=error_msg, name=name)

    def place(self,
              queries: Union[pd.DataFrame, Sequence[MolPlacementInput]],
              expand_isomers: bool = False,
              **kwargs) -> Union[pebble.ProcessMapFuture, pd.DataFrame]:
        """
        Due to the way Monster works merging A with B may yield a different result to B with A.
        Hence the ``permute`` boolean argument.
        """  # extended at end of file.
        if isinstance(queries, pd.DataFrame):
            assert 'smiles' in queries.columns
            assert 'name' in queries.columns
            assert 'hits' in queries.columns
            pre_iterator = queries.iterrows()
        elif isinstance(queries, Sequence):
            pre_iterator = enumerate(queries)
        else:  # it may crash...
            pre_iterator = queries

        def generator():
            for idx, data in pre_iterator:
                inputs = {'smiles': data['smiles'],
                       'name': data['name'],
                       'binary_hits': [binarize(m) for m in data['hits']]}
                if 'custom_map' in data:
                    inputs['custom_map'] = data['custom_map']
                if expand_isomers:
                    assert 'custom_map' not in inputs, 'custom_map not supported with expand_isomers'
                    for i, sub_smiles in enumerate(self.Victor.get_isomers_smiles(data['smiles'])):
                        inputs = {'name': data['name'] + f'-isomer_{i}',
                                  'smiles': sub_smiles}
                        yield inputs
                else:
                    yield inputs

        df = self(iterator=generator(), fun=self.place_subprocess, **kwargs)
        df['outcome'] = df.apply(functools.partial(self.categorize, size_tolerance=+50), axis=1)
        if 'unminimized_mol' in df.columns:
            df['unminimized_mol'] = df.unminimized_mol.fillna(Chem.Mol()) # noqa
        else:
            self.Victor.journal.critical('No unminimized_mol column (unsupported Victor?)')
            return df
        self.fix_intxns(df)  # noqa its in extras
        return df


# prepend docstring to combine
LabPlace.place.__doc__ = LabBench.__call__.__doc__ + '\n\n' + LabPlace.place.__doc__



"""
Due to the way ``py3Dmol`` works, I have to monkey patch it to add some functionality.
Namely, the ``py3Dmol.view.addModel`` is actually a __getattr__ call wrapping the JS,
which is not obeyed by super() and thus cannot be overridden by subclassing.

Therefore, I have to monkey patch it.

... code-block:: python

    viewer = py3Dmol.view()
    viewer.monkey_patch()
    viewer.add_template(vicky.apo_pdbblock)
    viewer.add_mol(vicky.hits[0], name='hit', carbon_color='cyan')
    viewer.add_mol(vicky.minimized_mol, name=vicky.long_name, colorscheme='coralCarbon')
    viewer.show()

carbon_color can be None (default), a CSS color name, or a hex color.
"""

# py3Dmol
import py3Dmol
from types import MethodType
from rdkit import Chem
from rdkit.Chem import AllChem
from typing import List, Optional

# https://3dmol.csb.pitt.edu/doc/global.html#:~:text=as%20the%20colorscheme-,Properties,-Name
py3Dmol.view.color_names = ['default',
                            'greenCarbon',
                            'cyanCarbon',
                            'magentaCarbon',
                            'purpleCarbon',
                            'whiteCarbon',
                            'orangeCarbon',
                            'yellowCarbon',
                            'blueCarbon']


def _add_mol(self, mol, name=None, carbon_color: Optional[str] = None, opacity=1., **kwargs):
    """
    Whereas there is ``IPythonConsole.addMolToView(caffeine, view)`` in rdkit using py3Dmol,
    I want more.

    :param mol:
    :param name:
    :param colorscheme:
    :param opacity:
    :param kwargs:
    :return:
    """
    if name is not None:
        pass
    elif mol.HasProp('_Name'):
        name: str = mol.GetProp('_Name')
    else:
        raise ValueError('No name provided and mol has no _Name property (title)')
    if name in self.model_addresses:
        raise ValueError(f'{name} is already present in the viewer')
    molblock: str = Chem.MolToMolBlock(mol)
    self.add_model(molblock, 'mol', name=name, **kwargs)
    # `.setStyle({'stick':{}})` will work, but I wanted the args to be explicit as I forget them.
    if carbon_color == 'default':
        self.setStyle({'model': -1, }, {'stick': {'colorscheme': 'default', 'opacity': opacity}})
    elif not isinstance(carbon_color, str):
        raise ValueError(f'No idea what is carbon_color={carbon_color}')
    elif 'Carbon' in carbon_color:
        self.setStyle({'model': -1, }, {'stick': {'colorscheme': carbon_color, 'opacity': opacity}})
    elif '#' not in carbon_color:
        self.setStyle({'model': -1, }, {'stick': {'colorscheme': f'{carbon_color}Carbon', 'opacity': opacity}})
    elif carbon_color == 'prop' or carbon_color is None:
        color = mol.GetProp('color')
        self.setStyle({'model': -1, }, {'stick': {'colorscheme': f'{color}Carbon', 'opacity': opacity}})
    else:
        # this will not work on an update which runs off updatejs
        self.startjs += f'''\n
    let customColorizeFor{name} = function(atom){{
          // attribute elem is from https://3dmol.csb.pitt.edu/doc/AtomSpec.html#elem
          if (atom.elem === 'C'){{
              return "{carbon_color}"
          }}else{{
              return $3Dmol.getColorFromStyle(atom, {{colorscheme: "whiteCarbon"}});
          }}
      }}\n'''
        self.setStyle({'model': -1}, {'stick': {'colorfunc': 'customColorize', 'opacity': opacity}})
        # make it a function not a string "customColorize"
        self.startjs = self.startjs.replace('"customColorize"', f'customColorizeFor{name}')


def _add_template(self, pdbblock: str, name='template', colorscheme='whiteCarbon'):
    self.add_model(pdbblock, 'PDB', name=name)
    self.setStyle({'model': -1}, {'cartoon': {'colorscheme': colorscheme, 'opacity': 1}})
    self.zoomTo()


def _get_idx(self, name):
    return self.model_addresses.index(name)


def _add_model(self, *args, name='unknown', **kwargs):
    self.addModel(*args, **kwargs)
    self.model_addresses.append(name)


def monkey_patch(viewer: py3Dmol.view):
    viewer.model_addresses: List[str] = []
    viewer.add_model = _add_model.__get__(viewer, py3Dmol.view)
    viewer.add_mol = _add_mol.__get__(viewer, py3Dmol.view)
    viewer.get_idx = _get_idx.__get__(viewer, py3Dmol.view)
    viewer.add_template = _add_template.__get__(viewer, py3Dmol.view)

py3Dmol.view.monkey_patch = monkey_patch

patched_3Dmol_view = py3Dmol.view


########################################################################################################################

__doc__ = \
    """
Combine => merge/join
This merge is used by all.

Peculiarly/for historical reasons the merging is done by fragmentation and joining the fragments 
as opposed to absorption, which is what happens in the ring expansion.

    """

########################################################################################################################

from typing import Optional, Dict, List, Union
from warnings import warn

from rdkit import Chem
from rdkit.Chem import rdmolops
from ._join_neighboring import _MonsterJoinNeigh
from .bond_provenance import BondProvenance
from .positional_mapping import GPM
from ..error import DistanceError, FragmensteinError


class _MonsterMerge(_MonsterJoinNeigh, GPM):

    def simply_merge_hits(self,
                          hits: Optional[List[Chem.Mol]] = None,
                          linked: bool = True,
                          ) -> Chem.Mol:
        """
        Recursively stick the hits together and average the positions.
        This is the monster of automerging, full-merging mapping and partial merging mapping.
        The latter however uses `partially_blend_hits` first.
        The hits are not ring-collapsed and -expanded herein.

        :param hits: optionally give a hit list, else uses the attribute ``.hits``.
        :param linked: if true the molecules are joined, else they are placed
            in the same molecule as disconnected fragments.
        :return: the rdkit.Chem.Mol object that will fill ``.scaffold``
        """
        if hits is None:
            hits = sorted(self.hits, key=lambda h: h.GetNumAtoms(), reverse=True)
        for hit in hits:
            BondProvenance.set_all_bonds(hit, 'original')
        self.journal.debug(f"Merging: {[hit.GetProp('_Name') for hit in hits]}")
        scaffold = Chem.Mol(hits[0])
        # first try
        save_for_later = []
        for fragmentanda in hits[1:]:
            try:
                scaffold = self.merge_pair(scaffold, fragmentanda)
            except FragmensteinError:
                save_for_later.append(fragmentanda)
        # second try
        join_later = []
        for fragmentanda in save_for_later:
            try:
                scaffold = self.merge_pair(scaffold, fragmentanda)
            except FragmensteinError:
                join_later.append(fragmentanda)
        # join (last ditch)
        for fragmentanda in join_later:
            if linked:
                try:
                    scaffold = self.join_neighboring_mols(scaffold, fragmentanda)
                except FragmensteinError as error:
                    self.unmatched.append(fragmentanda.GetProp("_Name"))
                    msg = f'Hit {fragmentanda.GetProp("_Name")} has no connections! Skipping!'
                    if self.throw_on_discard:
                        raise error
                    else:
                        warn(msg)
            else:
                new_name = self._get_combined_name(scaffold, fragmentanda)
                scaffold = rdmolops.CombineMols(scaffold, fragmentanda)
                scaffold.SetProp('_Name', str(new_name))
        return scaffold


    def merge_pair(self, scaffold: Chem.Mol, fragmentanda: Chem.Mol, mapping: Optional = None) -> Chem.Mol:
        """
        To specify attachments use ``.merge``.
        To understand what is going on see ``.categorize``

        :param scaffold: mol to be added to.
        :param fragmentanda: mol to be fragmented
        :param mapping: see ``get_positional_mapping``. Optional in _pre_fragment_pairs
        :return:
        """
        done_already = []
        fp = self._pre_fragment_pairs(scaffold, fragmentanda, mapping)
        # confusingly these are hit indexed.
        for anchor_index, attachment_details in fp.items():
            # anchor index is the fragment-to-added's internal atom that attaches
            if anchor_index in done_already:
                continue
            # fix rings.
            uniques = {atom.GetIdx() for atom in fragmentanda.GetAtoms() if
                       'overlapping' not in atom.GetProp('_Category')}
            team = self._recruit_team(fragmentanda, anchor_index, uniques)
            other_attachments = list((team & set(fp.keys())) - {anchor_index})
            other_attachment_details = []
            for other in other_attachments:
                other_attachment_details.append(fp[other])
                done_already.append(other)
            scaffold = self._merge_part(scaffold, fragmentanda,
                                        anchor_index=anchor_index,
                                        attachment_details=attachment_details,
                                        other_attachments=other_attachments,
                                        other_attachment_details=other_attachment_details)
        new_name = self._get_combined_name(scaffold, fragmentanda)
        scaffold.SetProp('_Name', str(new_name))
        self.keep_copy(scaffold, 'pair_merged')
        return scaffold

    def _get_combined_name(self, first_mol: Chem.Mol, second_mol: Chem.Mol):
        get_name = lambda mol: mol.GetProp('_Name') if mol.HasProp('_Name') else 'molecule'
        name_1 = get_name(first_mol)
        name_2 = get_name(second_mol)
        return f'{name_1}-{name_2}'

    # ------ merging by fragmentation  ------------------------

    def _pre_fragment_pairs(self, scaffold: Chem.Mol, fragmentanda: Chem.Mol, A2B_mapping: Optional = None) \
            -> Dict[int, List[Dict]]:
        """
        Returns

            {4: [{'idx': 5,
                   'type': rdkit.Chem.rdchem.BondType.SINGLE,
                   'idx_F': 5,
                   'idx_S': 1}], ...}

        which is slight more than {5: [{'idx': 4, 'type': rdkit.Chem.rdchem.BondType.SINGLE}], ... from categories

        idx_F: fragmentanda index
        idx_S: scaffold index

        required for self.merge, the key is the index of anchoring atom.

        Calls get_positional_mapping and _categorize.

        :param scaffold: mol to be added to.
        :param fragmentanda: mol to be fragmented
        :param A2B_mapping: see ``get_positional_mapping``
        :return:
        """
        # get A2B mapping
        if A2B_mapping is None:
            A2B_mapping = self.get_positional_mapping(scaffold, fragmentanda)
        get_key = lambda d, v: list(d.keys())[list(d.values()).index(v)]
        if len(A2B_mapping) == 0:
            raise DistanceError(hits=[scaffold, fragmentanda])
        # store alternative atom symbols.
        for si, fi in A2B_mapping.items():
            sa = scaffold.GetAtomWithIdx(si)
            sn = sa.GetSymbol()
            fn = fragmentanda.GetAtomWithIdx(fi).GetSymbol()
            if sn != fn:
                sa.SetProp('_AltSymbol', str(fn))
        # prepare.
        uniques = set(range(fragmentanda.GetNumAtoms())) - set(A2B_mapping.values())
        categories = self._categorize(fragmentanda, uniques)
        pairs = categories['pairs']
        for p in pairs:  # pairs:Dict[List[Dict]]
            for pp in pairs[p]:
                pp['idx_F'] = pp['idx']  # less ambiguous: fragmentanda index
                pp['idx_S'] = get_key(A2B_mapping, pp['idx'])  # scaffold index
        return pairs

    def _recruit_team(self, mol: Chem.Mol, starting: int, uniques: set, team: Optional[set] = None) -> set:
        if team is None:
            team = set()
        team.add(starting)
        for atom in mol.GetAtomWithIdx(starting).GetNeighbors():
            i = atom.GetIdx()
            if i in uniques and i not in team:
                team = self._recruit_team(mol, i, uniques, team)
        return team

    def _categorize(self, mol: Chem.Mol, uniques: set) -> Dict[str, Union[set, Dict]]:
        """
        What do the novel atoms do in terms of connectivity.
        Complicated dict output (called ``categories`` in the methods). Really ought to be SetProp of the atoms.

        * ``uniques`` are set of atoms to classify on
        * ``internals`` are unique atoms that are connected solely to unique atoms
        * ``attachments`` are non-unique atoms to which a unique atom connects
        * ``pairs`` is a dict of unique atom idx --> dict of ``idx`` --> attachment idx and ``type`` bond type.

        :param mol: molecule to describe
        :param uniques: set of indices that are new to this molecule
        :return:
        """
        #
        pairs = {}
        internals = set()
        attachments = set()
        dummies = set()
        for i in uniques:  # novel atoms
            unique_atom = mol.GetAtomWithIdx(i)
            if unique_atom.GetSymbol() == self.dummy_symbol:
                dummies.add(i)
            neighbours = {n.GetIdx() for n in unique_atom.GetNeighbors()}
            if len(neighbours - uniques) == 0:  # unlessone of the connections is not unique.
                internals.add(i)
            else:
                i_attached = neighbours - uniques
                attachments |= i_attached
                pairs[i] = [{'idx': j,
                             'type': mol.GetBondBetweenAtoms(i, j).GetBondType()} for j in i_attached]
        anchors = uniques - internals
        # store for safekeeping
        for atom in mol.GetAtoms():
            i = atom.GetIdx()
            if i in internals:  # novel and not connected
                atom.SetProp('_Category', 'internal')
            elif i in attachments:  # not-novel but connected
                atom.SetProp('_Category', 'overlapping-attachment')
            elif i in pairs:  # dict not set tho
                atom.SetProp('_Category', 'internal-attachment')
            else:  # overlapping
                atom.SetProp('_Category', 'overlapping')
        # if self._debug_draw: # depracated... but this could be useful...
        #     high = list(internals) + list(attachments) + list(anchors)
        #     color = {**{i: (0, 0.8, 0) for i in internals},
        #              **{i: (0, 0, 0.8) for i in attachments},
        #              **{i: (0.8, 0, 0.8) for i in anchors}}
        #     print('Purple: anchor atoms, Blue: attachments, Green: internals')
        #     self.draw_nicely(mol, highlightAtoms=high, highlightAtomColors=color)
        #     print({atom.GetIdx(): atom.GetProp('_Category') for atom in mol.GetAtoms()})
        return dict(uniques=uniques,
                    internals=internals,
                    attachments=attachments,
                    pairs=pairs,
                    dummies=dummies
                    )

    def _merge_part(self, scaffold: Chem.Mol, fragmentanda: Chem.Mol, anchor_index: int,
                    attachment_details: List[Dict],
                    other_attachments: List[int],
                    other_attachment_details: List[List[Dict]]) -> Chem.Mol:
        """
        This does the messy work for merge_pair.

        :param scaffold: the Chem.Mol molecule onto whose copy the fragmentanda Chem.Mol gets added
        :param fragmentanda: The other Chem.Mol molecule
        :param anchor_index: the fragment-to-added's internal atom that attaches (hit indexed)
        :param attachment_details: see `_pre_fragment_pairs` or example below fo an entry
        :type attachment_details: List[Dict]
        :param other_attachments:
        :param other_attachment_details:
        :return: a new Chem.Mol molecule

        Details object example:

            [{'idx': 5,
              'type': rdkit.Chem.rdchem.BondType.SINGLE,
              'idx_F': 5, # fragmentanda index
              'idx_S': 1  # scaffold index
              }], ...}
        """
        # get bit to add.
        bonds_to_frag = []
        for detail in attachment_details:
            attachment_index = detail['idx_F']  # fragmentanda attachment_index
            bonds_to_frag += [fragmentanda.GetBondBetweenAtoms(anchor_index, attachment_index).GetIdx()]
        bonds_to_frag += [fragmentanda.GetBondBetweenAtoms(oi, oad[0]['idx_F']).GetIdx() for oi, oad in
                          zip(other_attachments, other_attachment_details)]
        f = Chem.FragmentOnBonds(fragmentanda,
                                 bonds_to_frag,
                                 addDummies=False)
        frag_split = []
        fragmols = Chem.GetMolFrags(f, asMols=True, fragsMolAtomMapping=frag_split, sanitizeFrags=False)
        # Get the fragment of interest.
        ii = 0
        for mol_N, indices in enumerate(frag_split):
            if anchor_index in indices:
                break
            ii += len(indices)
        else:
            raise Exception
        frag = fragmols[mol_N]
        frag_anchor_index = indices.index(anchor_index)
        # pre-emptively fix atom ori_i
        # offset collapsed to avoid clashes.
        self.offset(frag)
        # Experimental code.
        # TODO: finish!
        # frag_atom = frag.GetAtomWithIdx(frag_anchor_index)
        # old2future = {atom.GetIntProp('_ori_i'): atom.GetIdx() + scaffold.GetNumAtoms() for atom in frag.GetAtoms()}
        # del old2future[-1] # does nothing but nice to double tap
        # if frag_atom.GetIntProp('_ori_i') == -1: #damn.
        #     for absent in self._get_mystery_ori_i(frag):
        #         old2future[absent] = scaffold_attachment_index
        # self._renumber_original_indices(frag, old2future)
        combo = Chem.RWMol(rdmolops.CombineMols(scaffold, frag))
        scaffold_anchor_index = frag_anchor_index + scaffold.GetNumAtoms()
        for detail in attachment_details:
            # scaffold_anchor_index : atom index in scaffold that needs to be added to scaffold_attachment_index
            # but was originally attached to attachment_index in fragmentanda.
            # the latter is not kept.
            attachment_index = detail['idx_F']  # fragmentanda attachment_index
            scaffold_attachment_index = detail['idx_S']  # scaffold attachment index
            bond_type = detail['type']
            combo.AddBond(scaffold_anchor_index, scaffold_attachment_index, bond_type)
            new_bond = combo.GetBondBetweenAtoms(scaffold_anchor_index, scaffold_attachment_index)
            # BondProvenance.set_bond(new_bond, '???')
            # self.transfer_ring_data(fragmentanda.GetAtomWithIdx(attachment_index),
            #                         combo.GetAtomWithIdx(scaffold_anchor_index))
        for oi, oad in zip(other_attachments, other_attachment_details):
            bond_type = oad[0]['type']
            scaffold_attachment_index = oad[0]['idx_S']
            scaffold_anchor_index = indices.index(oi) + scaffold.GetNumAtoms()
            combo.AddBond(scaffold_anchor_index, scaffold_attachment_index, bond_type)
            new_bond = combo.GetBondBetweenAtoms(scaffold_anchor_index, scaffold_attachment_index)
            # BondProvenance.set_bond(new_bond, '???')
        Chem.SanitizeMol(combo,
                         sanitizeOps=Chem.rdmolops.SanitizeFlags.SANITIZE_ADJUSTHS +
                                     Chem.rdmolops.SanitizeFlags.SANITIZE_SETAROMATICITY,
                         catchErrors=True)
        self._prevent_two_bonds_on_dummy(combo)
        scaffold = combo.GetMol()
        return scaffold

    def _prevent_two_bonds_on_dummy(self, mol: Chem.RWMol):
        """
        The case '*(C)C' is seen legitimately in some warheads... but in most cases these are not.

        :param mol:
        :return:
        """
        for atom in mol.GetAtoms():
            if atom.GetSymbol() != '*':
                pass
            elif len(atom.GetNeighbors()) <= 1:
                pass
            elif len(atom.GetNeighbors()) >= 2:
                self.journal.info(f'Dummy atom (idx={atom.GetIdx()}) has {len(atom.GetNeighbors())} bonds!')
                neighs = atom.GetNeighbors()
                first = neighs[0]
                for second in neighs[1:]:
                    rejected = second.GetIdx()  # that will be absorbed (deleted)
                    keeper = first.GetIdx()  # that absorbs (kept)
                    self._copy_bonding(mol, keeper, rejected)
                    self._mark_for_deletion(mol, rejected)
                self._delete_marked(mol)
                return self._prevent_two_bonds_on_dummy(mol)



from __future__ import annotations
# # original smaller function
# from collections import namedtuple
#
# Parts = namedtuple('Parts',
#                    ['step', 'headers', 'coordinates', 'connections', 'tails'],
#                    defaults=(0, [], [], [], [])
#                    )

# # The following does not work
#
# block = Chem.MolToPDBBlock(mol)
# mol = Chem.MolFromPDBBlock(block)
# for atom in mol.GetAtoms():
#     sn = atom.GetPDBResidueInfo().GetSerialNumber()
#     atom.GetPDBResidueInfo().SetSerialNumber(sn+10)
# block = Chem.MolToPDBBlock(mol)

from textwrap import wrap
import logging

log = logging.getLogger(__name__)

class MinimalPDBParser:
    """
    This purpose build PDB parser simply fixes the serial numbers.
    The reason is that writing a custom 50 line class is easier that
    having biopython or other non-builtin requirement as a requirement
    Importing the PDB into RDKit is inadvisable.
    """

    def __init__(self, block: str, remove_water=False, remove_other_hetatms=False, ligname="LIG"):
        self.ligname = ligname
        self.remove_water = remove_water
        self.remove_other_hetatms = remove_other_hetatms
        self.to_preserve_heteroResname = ["LIG"]
        self.water_resnames = ["HOH", "OH", "H"]
        if not remove_water:
            self.to_preserve_heteroResname +=  self.water_resnames

        self.step = 0
        # step = 0 header unfinished, 1 coordinates finished, 2 connections finished.
        self.headers = []
        self.coordinates = []
        self.connections = []
        self.tails = []
        self.parse(block)

    def parse(self, block:str) -> None:
        # ---- parse -----------------------------------------
        if block.find('\n') == -1:
            raise ValueError(f'No newline found in block: Is {block} a filename?')
        def starts_with(xrow, name): return xrow.find(name) == 0

        for row in block.split('\n'):
            row = row.strip()
            if row == '':
                continue
            elif starts_with(row, 'ATOM') or starts_with(row, 'HETATM'):
                if starts_with(row, 'HETATM'):
                    resType = row[17:21].strip()
                    if self.remove_other_hetatms:
                        if resType not in self.to_preserve_heteroResname:
                            continue
                    if self.remove_water:
                        if resType in self.water_resnames:
                            continue
                self.step = 1
                self.coordinates.append(row)
            elif starts_with(row, 'CONECT'):
                self.step = 2
                self.connections.append(row)
            elif starts_with(row, 'TER') or starts_with(row, 'END'):
                continue
            elif self.step == 0:
                self.headers.append(row)
            elif starts_with(row, 'ANISOU'):
                continue
            elif self.step == 1:
                log.info(f'What is {row}?')
            elif self.step > 1:
                self.tails.append(row)
            else:
                raise SyntaxError('Impossible')

    def __str__(self):
        return '\n'.join(self.headers + self.coordinates + self.connections + ['END'] + self.tails)

    def get_serial(self, entry: str) -> int:
        # ATOM    588 11 - 14
        return int(entry[6:12].strip())

    def get_residue_index(self, entry: str) -> int:
        # https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html
        # 23 - 26        Integer       resSeq       Residue sequence number.
        return int(entry[22:26].strip())

    def get_chain(self, entry: str) -> str:
        # 22             Character     chainID      Chain identifier.
        return entry[21].strip()

    def get_atomname(self, entry: str) -> str:
        # 13 - 16        Atom          name         Atom name.
        return entry[12:16].strip()

    def get_residue_name(self, entry: str) -> int:
        # https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html
        # 18 - 20        Residue name  resName      Residue name.
        return entry[17:20].strip()

    def get_max_serial(self) -> int:
        # assuming ordered
        return self.get_serial(self.coordinates[-1])

    def set_serial(self, entry: str, value: int) -> None:
        new = f'{entry[:6]}{value: >5}{entry[11:]}'
        i = self.coordinates.index(entry)
        self.coordinates[i] = new

    def offset_serials(self, offset: int) -> None:
        for entry in self.coordinates:
            original_serial = self.get_serial(entry)
            self.set_serial(entry, original_serial + offset)

    def offset_connections(self, offset:int) -> None:
        for i, entry in enumerate(self.connections):
            self.connections[i] = 'CONECT' + ''.join([f'{int(x)+offset: >5}' for x in wrap(entry[7:], 5)])

    def append(self, other: MinimalPDBParser):
        """
        Add a second parser data to it. But only its coordinates and connections.
        """
        offset = self.get_max_serial()
        other.offset_serials(offset)
        other.offset_connections(offset)
        self.coordinates += other.coordinates
        self.connections += other.connections

    def has_residue_index(self, index:int, chain: str):
        for entry in self.coordinates:
            if self.get_residue_index(entry) == index and self.get_chain(entry) == chain:
                return True
        else:
            return False

    def has_residue_name(self, name: str):
        """
        residue name, resn 3-letters
        """
        for entry in self.coordinates:
            if self.get_residue_name(entry) == name:
                return True
        else:
            return False





from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch

from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch

from ._refine import _MonsterRefine


class _MonsterFull(_MonsterRefine):
    # placement dependent methods

    # @classmethod #why was this a classmethod
    def full_blending(self) -> None:
        """
        a single scaffold is made (except for ``.unmatched``)
        """
        self.mol_options = [self.simply_merge_hits()]
        scaffold = self.posthoc_refine(self.mol_options[0])
        chimera = self.make_chimera(scaffold)

        # atom_map is filled via `self.get_mcs_mapping(target_mol, template_mol)`:
        self.positioned_mol = self.place_from_map(target_mol=self.initial_mol,
                                                  template_mol=chimera,
                                                  atom_map=None,
                                                  random_seed=self.random_seed)
        self.keep_copy(scaffold, 'scaffold')
        self.keep_copy(chimera, 'chimera')


import unittest

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem
from typing import List
from fragmenstein import Walton

# ======================================================================================================================

class WaltonTests(unittest.TestCase):

    def smiles_assertEqual(self, a, b):
        """
        helper method to test equality of smiles.
        Simply reshuffled the parts. Lazy.
        """
        if isinstance(a, str):
            expected_smiles = a
            mol = b
        else:
            expected_smiles = b
            mol = a
        obtained_smiles = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(expected_smiles, obtained_smiles)

    def test_superpose_map(self):
        demo = Walton.from_smiles(resorcinol='c1ccc(O)cc1O', eugenol='Oc1ccc(cc1OC)CC=C')  # create instance
        demo.superpose_by_map({(0, 1): {4: 0, 3: 1, 2: 2}})  # superpose molecules by atom indices
        demo()  # merge (Fragmenstein's Monster)
        self.smiles_assertEqual('C=CCc1ccc(O)c(OC)c1O', demo.merged)

    def test_polygon(self):
        benzene = Walton.create_polygon(n=6, bond_type=Chem.BondType.AROMATIC)
        self.smiles_assertEqual('c1ccccc1', benzene)

    def test_ring_on_plane(self):
        demo = Walton.from_smiles(furan='c1ccco1')
        demo.ring_on_plane(mol_idx=0)  # flat on xy
        demo.atom_on_axis(mol_idx=0, atom_idx=4, axis='x')  # ox on axis
        # the atom_idx can be actually be passed a Geometry.Point3D:
        demo.atom_on_axis(mol_idx=0,
                          atom_idx=demo.get_centroid_of_atoms(1, 2, mol_idx=0),
                          axis='x')
        for i in range(demo.mols[0].GetNumAtoms()):
            self.assertAlmostEqual(demo.get_point(i, 0).z, 0, -1)

    def pull_apart(self, mols: List[Chem.Mol], distance: float) -> Chem.Mol:
        walton = Walton(mols)
        walton.ring_on_plane(ring_idx=0, mol_idx=0)
        walton.superpose_by_mcs()
        walton.translate_parallel(mol_idx=1, distance=distance,
                                  base_atom_idx=0, pointer_atom_idx=2)
        walton(joining_cutoff=10)
        return walton.merged

    def test_pull_apart(self):
        """
        Two benzene molecules are placed at different distances.
        """
        benzene = Chem.MolFromSmiles('c1ccccc1')
        benzene.SetProp('_Name', 'benzene')
        AllChem.EmbedMolecule(benzene)
        expectations = {0.0: 'c1ccccc1',  # benzene
                        2.5: 'c1ccc2ccccc2c1',  # naphthalene
                        #3.5: 'C1CCC2(CC1)CCCCC2',  # spiro... inconsistent!
                        4.0: 'c1ccc(-c2ccccc2)cc1', # bonded
                        4.5: 'c1ccc(Oc2ccccc2)cc1',  # oxy bridge
                        6.0: 'c1ccc(COc2ccccc2)cc1',  # methoxy bridge
                        7.0: 'c1ccc(CCOc2ccccc2)cc1',  # ethoxy bridge
                        8.0: 'c1ccc(OCCOc2ccccc2)cc1', # glycol bridge
                        }
        for distance, expected in expectations.items():
            print(distance)
            merged: Chem.Mol = self.pull_apart([Chem.Mol(benzene), Chem.Mol(benzene)], distance)
            self.smiles_assertEqual(expected, merged)

    def test_tetrahedron(self):
        # build a tetrahedron!
        walton = Walton.from_smiles(ethanol='CCO', ethylamine='NCC')
        for i in range(2):
            walton.flatten_trio(mol_idx=i, atom_idcs=(0, 1, 2))
            walton.atom_to_origin(mol_idx=i, atom_idx=1)
        walton.rotate(mol_idx=1, theta=180, axis='z')
        walton.rotate(mol_idx=1, theta=90, axis='y')
        walton()
        self.smiles_assertEqual('CC(C)(N)O', walton.merged)

    def test_tetraherdron_with_atom_change(self):
        walton = Walton.from_smiles(methylether='COC', methylether2='COC')
        for i in range(2):
            walton.flatten_trio(mol_idx=i, atom_idcs=(0, 1, 2))
            walton.atom_to_origin(mol_idx=i, atom_idx=1)
        walton.rotate(mol_idx=i, theta=180, axis='z')
        walton.rotate(mol_idx=i, theta=90, axis='y')
        walton()
        self.smiles_assertEqual('CC(C)(C)C', walton.merged)


if __name__ == '__main__':
    unittest.main()



from rdkit import Chem
from ..victor import Victor
from ..m_rmsd import mRMSD
import os, json
from rdkit_to_params import Params
from rdkit import Chem
from rdkit.Chem import AllChem

class Wictor(Victor):
    """
    This Victor does not call Igor

    ... code-block:: python
        from framgenstein.faux_victors import Wictor
        from fragmenstein.demo import Mac1

        Wictor.capture_rdkit_log()
        # Wictor.enable_stdout()
        Wictor.error_to_catch = ()
        wicky = Wictor(hits=Mac1.get_n_filtered_mols(2),
                       pdb_block=Mac1.get_template(),
                       )
        wicky.combine()
    """
    uses_pyrosetta = False

    def _process_settings(self):
        self.journal.debug('Valid settings.py: ff_max_displacement, ff_constraint=1, ff_max_iterations')

    def _calculate_combination_thermo(self):
        # override igor.
        self._calculate_thermo_common()
        # save to disc
        self._checkpoint_charlie()

    def _calculate_placement_thermo(self):
        # override igor.
        self._calculate_thermo_common()
        # save to disc
        self._checkpoint_charlie()

    def _checkpoint_charlie(self):
        # making folder.
        self.make_output_folder()
        # no igor!
        self._log_warnings()
        self.journal.debug(f'{self.long_name} - saving pose collage')
        min_file = os.path.join(self.work_path, self.long_name, self.long_name + '.holo_minimised.pdb')
        with open(min_file, 'w') as w:
            w.write(self.minimized_pdbblock)
        self.journal.debug(f'{self.long_name} - saving Gibbs')
        # recover bonds
        lig_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.mol')
        Chem.MolToMolFile(self.minimized_mol, lig_file)
        score_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.json')
        with open(score_file, 'w') as w:
            json.dump({'Energy': self.energy_score,
                       'mRMSD': self.mrmsd.mrmsd,
                       'RMSDs': self.mrmsd.rmsds}, w)
        self._log_warnings()

    def _calculate_thermo_common(self):
        """
        This method is common to both combination and placement and is unique to Wictor
        as both ``_calculate_combination_thermo`` and ``_calculate_placement_thermo`` are overridden
        and do the same here.
        """
        # in _calculate_*_chem this was set:
        # self.mol = self.monster.positioned_mol
        # I need to assign atom names and stuff still:
        self.params = Params.from_mol(self.mol, name=self.ligand_resn, generic=True)
        self.params.NAME = self.ligand_resn  # force it.
        self.params.polish_mol()
        self.params.comments.clear()
        self.params.comments.append('Generated via Fragmenstein')
        if self.settings['ff_use_neighborhood']:  # default True
            neighborhood = self.monster.get_neighborhood(self.apo_pdbblock,
                                                         cutoff=self.settings['ff_neighborhood'],
                                                         addHs=True)
        else:
            neighborhood = None
        # allow_lax reduces the constraints if it fails
        min_result = self.monster.mmff_minimize(self.mol,
                                                neighborhood=neighborhood,
                                                ff_max_displacement=float(self.settings['ff_max_displacement']), # def 0
                                                ff_constraint=int(self.settings['ff_constraint']), # def 10
                                                ff_max_iterations=int(self.settings['ff_max_iterations']), # def 200
                                                allow_lax=bool(self.settings['ff_prevent_cis']), # true
                                                prevent_cis=bool(self.settings['ff_prevent_cis']), # true
                                                )
        self.minimized_mol: Chem.Mol = min_result.mol
        self.minimized_pdbblock: str = self._plonk_monster_in_structure(prepped_mol=self.minimized_mol)
        # The ddG is how strained the molecule is out of the protein... not the drop from binding.
        # recalculating:
        # min_result.ideal is with ff_minimise_ideal True
        ideal: Chem.Mol = self.monster.make_ideal_mol(ff_minimise=bool(self.settings['ff_minimise_ideal']))
        if neighborhood: # option ``ff_use_neighborhood`` is False. Why one would do this?
            AllChem.SanitizeMol(neighborhood)
        ideal_E: float = float('nan')
        ligand_E: float = float('nan')
        holo_E: float = float('nan')
        apo_E: float = float('nan')
        if 'Energy' not in self.minimized_mol.GetPropNames():
            self.ddG = float('nan')
            self.mrmsd: mRMSD = self._calculate_rmsd()
        elif neighborhood is None:
            ideal_E = ideal.GetDoubleProp('Energy')
            ligand_E = self.minimized_mol.GetDoubleProp('Energy')
        else:
            ideal_E = ideal.GetDoubleProp('Energy')
            ligand_E = self.minimized_mol.GetDoubleProp('Energy')
            # The holo needs recalculating as I don't want the constraints
            holo_E = self.monster.MMFF_score(Chem.CombineMols(self.minimized_mol, neighborhood), delta=False)
            apo_E = self.monster.MMFF_score(neighborhood, delta=False)
        # store data:
        self.energy_score['ideal'] = dict(total_score=ideal_E, unit='kcal/mol')
        self.energy_score['insitu'] = dict(total_score=ligand_E, unit='kcal/mol')
        self.energy_score['bound'] = dict(total_score=holo_E, unit='kcal/mol')
        self.energy_score['unbound'] = dict(total_score=apo_E + ideal_E, unit='kcal/mol')
        self.energy_score['apo'] = dict(total_score=apo_E, unit='kcal/mol')
        self.ddG: float = holo_E - apo_E - ideal_E
        self.mrmsd: mRMSD = self._calculate_rmsd()



from ..victor import Victor
from .fritz import Fritz
from ..error import FragmensteinError
from rdkit import Chem
from rdkit.Chem import AllChem
from typing import List, Optional, Dict, Union, Callable
import os, json, time

class OpenVictor(Victor):
    uses_pyrosetta = False

    def _calculate_combination_thermo(self):
        return self._calculate_thermo()

    def _calculate_placement_thermo(self):
        return self._calculate_thermo()

    def _calculate_thermo(self):
        if self.is_covalent:
            raise NotImplementedError('OpenVictor does not support covalent ligands')
        self.journal.debug(f'{self.long_name} - Starting system setup')
        # in _calculate_*_chem this was set:
        # self.mol = self.monster.positioned_mol
        restraint_k = self.settings['mm_restraint_k'] # default 1000.
        tolerance = self.settings['mm_tolerance']  # 10 * mmu.kilocalorie_per_mole / (mmu.nano * mmu.meter)
        maxIterations = self.settings['mm_max_iterations'] # 0 is infinite
        self.fritz = Fritz(prepped_mol=self.preminimized_undummied_mol,
                           pdb_block=self.apo_pdbblock,
                           resn=self.ligand_resn,
                           restraining_atom_indices=self._get_restraining_atom_indices(),
                           restraint_k=restraint_k,
                           mobile_radius=self.settings['mm_mobile_radius'],
                           )
        self.unminimized_pdbblock = self.fritz.to_pdbblock()
        self._data: Dict = self.fritz.reanimate(tolerance=tolerance, maxIterations=maxIterations)
        while not self.quick_reanimation and self._data['binding_dG'] > 0 * self.fritz.molar_energy_unit:
            restraint_k = 2.
            self.fritz.alter_restraint(restraint_k / 3.)
            self._data: Dict = self.fritz.reanimate(tolerance=tolerance, maxIterations=maxIterations)
            self.journal.debug(f'{self.long_name} - restraints at {restraint_k}')
        self._data['restraint_k'] = restraint_k
        self._data['origins'] = self.monster.origin_from_mol()
        self.energy_score = {k: v.value_in_unit(self.fritz.molar_energy_unit)
                                    if hasattr(v, 'value_in_unit') else v for k, v in self._data.items()
                                                                 if k not in ('minimized_pdb', 'origins')}
        self.energy_score['unit'] = str(self.fritz.molar_energy_unit)
        self.minimized_pdbblock = self.fritz.to_pdbblock()
        self.minimized_mol = self.fritz.to_mol()
        self.mrmsd = self._calculate_rmsd()
        self.checkpoint()
        self.tock = time.time()
        return self.summarize()

    def _process_settings(self):
        pass

    def _get_restraining_atom_indices(self) -> List[int]:
        # Place has '_Stdev': 1.887379141862766e-15, '_Origin': '["x0395.9", "x0434.6"]', '_Max': 0.37818712299601354}
        # Combine has '_ori_i': 100, '_ori_name': 'x0395', '_x': 9.309, '_y': -5.402, '_z': 26.27
        # and occassionally '_ring_i'
        # The problem stemming from the collasing business
        # for now, all atoms are the same.
        restrainables: List[int] = []
        for atom in self.monster.positioned_mol.GetAtoms():
            if atom.GetSymbol() == 'H':
                continue
            elif atom.HasProp('_ori_name') or atom.HasProp('_Origin'):
                restrainables.append( atom.GetIdx() )
            else:
                # no restraint
                pass
        return restrainables

    def checkpoint(self, save_outputs: Optional[bool] = None):
        if save_outputs is True:
            pass
        elif self.settings.get('save_outputs', False):
            return
        self.journal.debug(f'{self.long_name} - saving data to disk at {self.work_path}')
        self.make_output_folder()
        with open(os.path.join(self.work_path, self.long_name, self.long_name + '.holo_unminimised.pdb'), 'w') as w:
            w.write(self.unminimized_pdbblock)
        with open(os.path.join(self.work_path, self.long_name, self.long_name + '.holo_minimised.pdb'), 'w') as w:
            w.write(self.minimized_pdbblock)
        for hit in self.hits:
            Chem.MolToMolFile(hit, os.path.join(self.work_path, self.long_name, hit.GetProp('_Name') + '.mol'))
        Chem.MolToMolFile(self.monster.positioned_mol,
                          os.path.join(self.work_path, self.long_name, self.long_name + '.positioned.mol'))
        Chem.MolToMolFile(self.fritz.prepped_mol,
                          os.path.join(self.work_path, self.long_name, self.long_name + '.prepped.mol'))
        Chem.MolToMolFile(self.minimized_mol,
                          os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.mol'))
        with open(os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.json'), 'w') as w:
            json.dump({**self.energy_score, 'origins': self._data['origins']}, w)

    def summarize(self):
        if self.error_msg:
            if self.monster is None:
                N_constrained_atoms = float('nan')
                N_unconstrained_atoms = float('nan')
            elif self.monster.positioned_mol is None:
                N_constrained_atoms = float('nan')
                N_unconstrained_atoms = float('nan')
            else:
                N_constrained_atoms = self.constrained_atoms
                N_unconstrained_atoms = self.unconstrained_heavy_atoms
            return {'name': self.long_name,
                    'smiles': self.smiles,
                    'error': self.error_msg,
                    'mode': self.merging_mode,
                    '∆∆G': float('nan'),
                    '∆G_bound': float('nan'),
                    '∆G_unbound': float('nan'),
                    'comRMSD': float('nan'),
                    'N_constrained_atoms': N_constrained_atoms,
                    'N_unconstrained_atoms': N_unconstrained_atoms,
                    'runtime': self.tock - self.tick,
                    'regarded': self.monster.matched,
                    'disregarded': self.monster.unmatched
                    }
        else:
            return {'name': self.long_name,
                    'smiles': self.smiles,
                    'error': self.error_msg,
                    'mode': self.merging_mode,
                    '∆∆G': self.energy_score.get('binding_dG', float('nan')),
                    'comRMSD': self.mrmsd.mrmsd,
                    'N_constrained_atoms': self.constrained_atoms,
                    'N_unconstrained_atoms': self.unconstrained_heavy_atoms,
                    'runtime': self.tock - self.tick,
                    'regarded': self.monster.matched,
                    'disregarded': self.monster.unmatched,
                    'origins': self._data['origins'],
                    }

    def _fix_minimized(self, *args, **kwargs):
        raise NotImplementedError('OpenVictor does not support covalent ligands')



"""
This is an optional import and can only be accessed via `fragmenstein.laboratory.validator`.

For place a route is providing a dataframe with the following columns:

- ``name``: the name of the built molecule (_e.g._'build#👾')
- ``smiles``: the smiles of the built molecule
- ``hits``: the hits used to build the molecule

These can be passed to ``Laboratory.place`` to place the built molecules.
This is a pandera schema for validation.

.. code-block:: python
    from fragmenstein.laboratory.validator import place_input_validator
    place_input_validator.validate(df)

    from fragmenstein import Laboratory
    Laboratory(pdbblock=👾👾👾, covalent_resi=None).place(df, expand_isomers=False, n_cores=12)
"""

import pandas as pd
import pandera as pa
from rdkit import Chem


def hits_check(mols_col: pd.Series) -> bool:
    for mols in mols_col:
        if not isinstance(mols, list):
            return False
        for mol in mols:
            if not isinstance(mol, Chem.Mol):
                return False
    return True


def smiles_check(smiles_col: pd.Series) -> bool:
    for smiles in smiles_col:
        if Chem.MolFromSmiles(smiles) is None:
            return False
    return True


place_input_validator = pa.DataFrameSchema({
    "name": pa.Column(pa.String),
    "hits": pa.Column(checks=[pa.Check(hits_check)]),
    "smiles": pa.Column(pa.String, checks=[pa.Check(smiles_check)])})



import logging
import unittest
import tempfile
from fragmenstein.demo import Mac1
from fragmenstein import Laboratory
import pandas as pd
import os

Laboratory.Victor.enable_stdout(logging.CRITICAL)

class LabTests(unittest.TestCase):
    def test_lab_combine(self):
        pdb_block = Mac1.get_template()
        hits = [Mac1.get_mol(f'diamond-{name}') for name in ['x0282_A','x0104_A','x0722_A','x0591_A','x0091_B']]
        lab = Laboratory(pdbblock=pdb_block, covalent_resi=None, run_plip=True)
        combinations: pd.DataFrame = lab.combine(hits, n_cores=8)
        self.assertGreater(len(combinations.outcome == 'acceptable'), 0)






########################################################################################################################

__doc__ = \
    """
Base metho=ods
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2020 A.D."
__license__ = "MIT"
__citation__ = ""

from ..version import __version__

########################################################################################################################


import logging, warnings
import re
import os
import time
from typing import List, Union, Optional, Callable, Dict

from rdkit import Chem
from ..m_rmsd import mRMSD
from ..monster import Monster   # will become Victor.Monster
from ..settings import default_settings, default_settings_yaml


class _VictorBase:
    uses_pyrosetta = True
    quick_reanimation = False  # thorugh reanimation?
    # These will be depraecated
    monster_average_position = default_settings['monster_average_position']  # default False
    monster_throw_on_discard = default_settings['monster_throw_on_discard']  # default False
    monster_mmff_minisation = default_settings['ff_minisation']  # default True
    # These will likely stay
    constraint_function_type = 'FLAT_HARMONIC'
    work_path = default_settings['work_path']
    journal = logging.getLogger('Fragmenstein')
    journal.setLevel(logging.DEBUG)
    # here for ease of subclassing
    Monster = Monster

    covalent_definitions = [{'residue': 'CYS', 'smiles': '*SC', 'atomnames': ['CONN3', 'SG', 'CB']}]
    warhead_definitions = [{'name': 'acrylamide',
                            'covalent': 'C(=O)CC*',  # the N may be secondary etc. so best not do mad substitutions.
                            'covalent_atomnames': ['CZ', 'OZ', 'CY', 'CX', 'CONN1'],
                            # OZ needs to tautomerise & h-bond happily.
                            'noncovalent': 'C(=O)C=C',
                            'noncovalent_atomnames': ['CZ', 'OZ', 'CY', 'CX']},
                           {'name': 'chloroacetamide',
                            'covalent': 'C(=O)C*',  # the N may be secondary etc. so best not do mad substitutions.
                            'covalent_atomnames': ['CY', 'OY', 'CX', 'CONN1'],
                            # OY needs to tautomerise & h-bond happily.
                            'noncovalent': 'C(=O)C[Cl]',
                            'noncovalent_atomnames': ['CY', 'OY', 'CX', 'CLX']
                            },
                           {'name': 'nitrile',
                            'covalent': 'C(=N)*',  # zeroth atom is attached to the rest
                            'covalent_atomnames': ['CX', 'NX', 'CONN1'],
                            'noncovalent': 'C(#N)',  # zeroth atom is attached to the rest
                            'noncovalent_atomnames': ['CX', 'NX']
                            },
                           {'name': 'vinylsulfonamide',
                            'covalent': 'S(=O)(=O)CC*',  # the N may be secondary etc. so best not do mad substitutions.
                            'covalent_atomnames': ['SZ', 'OZ1', 'OZ2', 'CY', 'CX', 'CONN1'],  # OZ tauto
                            'noncovalent': 'S(=O)(=O)C=C',
                            'noncovalent_atomnames': ['SZ', 'OZ1', 'OZ2', 'CY', 'CX']
                            },
                           {'name': 'bromoalkyne',
                            'covalent': 'C(=C)*',
                            'covalent_atomnames': ['CX', 'CY', 'CONN1'],
                            # OY needs to tautomerise & h-bond happily.
                            'noncovalent': 'C#C[Br]',
                            'noncovalent_atomnames': ['CX', 'CY', 'BRX']
                            },
                           ]

    # these may be wrong and need checking.
    possible_definitions = [{'name': 'aurothiol',  # gold salt
                             'covalent': 'S[Au]*',
                             'covalent_atomnames': ['SY', 'AUX', 'CONN1'],
                             # OY needs to tautomerise & h-bond happily.
                             'noncovalent': 'S[Au]P(CC)(CC)CC',
                             'noncovalent_atomnames': ['SY', 'AUX', 'PL', 'CL1', 'CL2', 'CL3', 'CL4', 'CL5', 'CL6']
                             },
                            {'name': 'aldehyde',
                             'covalent': 'C(O)*',
                             'covalent_atomnames': ['CX', 'OX', 'CONN1'],
                             'noncovalent': '[C:H1]=O',  # this at
                             'noncovalent_atomnames': ['CX', 'OX']
                             },
                            ]

    _connected_names = ('CONN', 'LOWE', 'UPPE', 'CONN1', 'CONN2', 'CONN3', 'LOWER', 'UPPER')

    error_to_catch = ()  # Exception

    remove_other_hetatms = True  # remove all protein heteroatoms that are not water or ligand when loaded from PDB

    # ================== Init ==========================================================================================

    def __init__(self,
                 hits: List[Chem.Mol],
                 pdb_filename: Union[None, str] = None,
                 pdb_block: Union[None, str] = None,
                 ligand_resn: str = 'LIG',
                 ligand_resi: Union[int, str, None] = None,
                 covalent_resn: str = 'CYS',  # no other option is accepted.
                 covalent_resi: Optional[Union[int, str]] = None,
                 extra_protein_constraint: Union[str] = None,
                 pose_fx: Optional[Callable] = None,
                 monster_random_seed: Optional[int] = None,
                 **settings
                 ) -> object:
        """
        Initialise Victor in order to allow either combinations (merging/linking without a given aimed for molecule)
        or placements (using a given aimed for molecule).

        :param hits: list of rdkit molecules
        :param pdb_filename: file of apo structure
        :param pdb_block: alternative for above: a string of apo structure
        :param ligand_resn: 3 letter code or your choice
        :param ligand_resi: Rosetta-style pose(int) or pdb(str)
        :param covalent_resn: only CYS accepted. if smiles has no * it is ignored
        :param covalent_resi: Rosetta-style pose(int) or pdb(str)
        :param extra_protein_constraint: multiline string of constraints relevant to the protein
        :param pose_fx: a function to call with pose to tweak or change something before minimising.
        :param monster_random_seed: a random seed for rdkit Embedding
        :param settings: Not used in base version of Victor
        """
        # ## Store
        # entry attributes
        if pdb_filename:
            with open(pdb_filename) as fh:
                self.apo_pdbblock = fh.read()
        elif pdb_block:
            self.apo_pdbblock = pdb_block
        else:
            raise ValueError('Provide a pdb_filename or pdb_block of the template')
        self.hits = hits
        self.ligand_resn: str = ligand_resn.upper()
        self.ligand_resi: str = ligand_resi if ligand_resi else self._get_empty_resi()
        self.covalent_resn = covalent_resn.upper()
        self.covalent_resi = covalent_resi
        self._correct_covalent_resi()  # noqa defined in plonk. todo: split into covalent and anchor residue.
        self.extra_constraint = extra_protein_constraint
        self.pose_fx = pose_fx
        self.random_seed = monster_random_seed
        self.settings = {**default_settings, **settings}
        self._process_settings()
        # this is readied in case user wants to change it:
        self.monster = self.Monster(hits,
                                    average_position=self.monster_average_position,
                                    random_seed=self.random_seed)
        self.monster.throw_on_discard = self.monster_throw_on_discard
        self.monster.linker_element = self.settings.get('linker_element', 'N')
        # ## Fill by place and combine differently
        self.long_name: str = 'ligand'
        self.smiles = None
        # ## Filled by place
        self.merging_mode = "not_set"
        self.custom_map = {}  # this will be fixed in monster by `.fix_custom_map`
        # ## Filled by combine
        self.joining_cutoff = None
        # ## Calculated
        self.is_covalent = None
        self.params = None  # this will be the rdkit_to_params.Params instance
        self.mol = None   # this will be the unminimised mol
        self.constraint = None  # Rosetta string-form constraint defs
        self.modifications = {}
        self.unminimized_pdbblock = None
        self.igor = None
        self.unbound_pose = None
        self.minimized_pdbblock = None
        self.minimized_mol = None
        self.reference_mol = None  # filled only for validate
        # buffers etc.
        self._warned = warnings.catch_warnings()  # new one will be made. here for clarify
        self.energy_score = {'bound': {'total_score': float('nan')},
                             'unbound': {'total_score': float('nan')}}
        self.mrmsd = mRMSD.mock()
        self.ddG = float('nan')
        # for debug purposes
        self.tick = time.time()
        self.tock = float('inf')
        self.error_msg = ''

    # ----------------- init called methods ----------------------------------------------------------------------------
    @classmethod
    def slugify(cls, name: str):
        return re.sub(r'[\W_.-]+', '-', name)

    def _process_settings(self):
        # what settings.py are valid?
        pass

_VictorBase.__init__.__doc__ += '\n Some arguments can be defined externally: \n'+default_settings_yaml



from __future__ import annotations
########################################################################################################################

__doc__ = \
    """
These are extras.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2020 A.D."
__license__ = "MIT"
__version__ = "0.4"
__citation__ = ""

########################################################################################################################

import logging
import os
import re
import requests
import sys, json
import unicodedata
import numpy as np
from ..extraction_funs import add_dummy_to_mol
from typing import List, Union, Optional, Dict, Tuple, TYPE_CHECKING

from rdkit import Chem
from rdkit.Chem import rdFMCS, AllChem, EnumerateStereoisomers
from rdkit.Chem.MolStandardize import rdMolStandardize

from ._victor_common import _VictorCommon
from ._victor_show import _VictorShow
from ..m_rmsd import mRMSD
from ..monster import Monster
from rdkit_to_params import Params
from ..igor import Igor

if TYPE_CHECKING or 'sphinx' in sys.modules:
    import pandas as pd


class _VictorUtils(_VictorShow):  # _VictorCommon -> _VictorShow

    def dock(self) -> Chem.Mol:
        """
        The docking is done by ``igor.dock()``. This basically does that, extacts ligand, saves etc.

        :return:
        """
        docked = self.igor.dock()
        self.docked_pose = docked
        docked.dump_pdb(f'{self.work_path}/{self.long_name}/{self.long_name}.holo_docked.pdb')
        ligand = self.igor.mol_from_pose(docked)
        template = AllChem.DeleteSubstructs(self.params.mol, Chem.MolFromSmiles('*'))
        lig_chem = AllChem.AssignBondOrdersFromTemplate(template, ligand)
        lig_chem.SetProp('_Name', 'docked')
        Chem.MolToMolFile(lig_chem, f'{self.work_path}/{self.long_name}/{self.long_name}.docked.mol')
        return lig_chem
        # print(pyrosetta.get_fa_scorefxn()(docked) - v.energy_score['unbound']['total_score'])

    def summarize(self):
        if self.error_msg:
            if self.monster is None:
                N_constrained_atoms = float('nan')
                N_unconstrained_atoms = float('nan')
            elif self.monster.positioned_mol is None:
                N_constrained_atoms = float('nan')
                N_unconstrained_atoms = float('nan')
            else:
                N_constrained_atoms = self.constrained_atoms
                N_unconstrained_atoms = self.unconstrained_heavy_atoms
            return {'name': self.long_name,
                    'smiles': self.smiles,
                    'error': self.error_msg,
                    'mode': self.merging_mode,
                    '∆∆G': float('nan'),
                    '∆G_bound': float('nan'),
                    '∆G_unbound': float('nan'),
                    'comRMSD': float('nan'),
                    'N_constrained_atoms': N_constrained_atoms,
                    'N_unconstrained_atoms': N_unconstrained_atoms,
                    'runtime': self.tock - self.tick,
                    'regarded': self.monster.matched,
                    'disregarded': self.monster.unmatched
                    }
        else:
            return {'name': self.long_name,
                    'smiles': self.smiles,
                    'error': self.error_msg,
                    'mode': self.merging_mode,
                    '∆∆G': self.energy_score['bound']['total_score'] - \
                           self.energy_score['unbound']['total_score'],
                    '∆G_bound': self.energy_score['bound']['total_score'],
                    '∆G_unbound': self.energy_score['unbound']['total_score'],
                    'comRMSD': self.mrmsd.mrmsd,
                    'N_constrained_atoms': self.constrained_atoms,
                    'N_unconstrained_atoms': self.unconstrained_heavy_atoms,
                    'runtime': self.tock - self.tick,
                    'regarded': self.monster.matched,
                    'disregarded': self.monster.unmatched
                    }

    # =================== Other ========================================================================================

    @classmethod
    def copy_names(cls, acceptor_mol: Chem.Mol, donor_mol: Chem.Mol):
        """
        Copy names form donor to acceptor by finding MCS.
        Does it properly and uses ``PDBResidueInfo``.

        :param acceptor_mol: needs atomnames
        :param donor_mol: has atomnames
        :return:
        """
        mcs = rdFMCS.FindMCS([acceptor_mol, donor_mol],
                             atomCompare=rdFMCS.AtomCompare.CompareElements,
                             bondCompare=rdFMCS.BondCompare.CompareOrder,
                             ringMatchesRingOnly=True)
        common = Chem.MolFromSmarts(mcs.smartsString)
        pos_match = acceptor_mol.positioned_mol.GetSubstructMatch(common)
        pdb_match = donor_mol.GetSubstructMatch(common)
        for m, p in zip(pos_match, pdb_match):
            ma = acceptor_mol.GetAtomWithIdx(m)
            pa = donor_mol.GetAtomWithIdx(p)
            assert ma.GetSymbol() == pa.GetSymbol(), 'The indices do not align! ' + \
                                                     f'{ma.GetIdx()}:{ma.GetSymbol()} vs. ' + \
                                                     f'{pa.GetIdx()}:{pa.GetSymbol()}'
            ma.SetMonomerInfo(pa.GetPDBResidueInfo())

    @classmethod
    def add_constraint_to_warhead(cls, name: str, constraint: str):
        """
        Add a constraint (multiline is fine) to a warhead definition.
        This will be added and run by Igor's minimiser.

        :param name:
        :param constraint:
        :return: None
        """
        for war_def in cls.warhead_definitions:
            if war_def['name'] == name:
                war_def['constraint'] = constraint
                break
        else:
            raise ValueError(f'{name} not found in warhead_definitions.')

    @classmethod
    def distance_hits(cls, pdb_filenames: List[str],
                      target_resi: int,
                      target_chain: str,
                      target_atomname: str,
                      ligand_resn='LIG') -> List[float]:
        """
        See closest hit for info.

        :param pdb_filenames:
        :param target_resi:
        :param target_chain:
        :param target_atomname:
        :param ligand_resn:
        :return:
        """
        distances = []
        import pymol2
        with pymol2.PyMOL() as pymol:
            for hit in pdb_filenames:
                pymol.cmd.load(hit)
                distances.append(min(
                    [pymol.cmd.distance(f'chain {target_chain} and resi {target_resi} and name {target_atomname}',
                                        f'resn {ligand_resn} and name {atom.name}') for atom in
                     pymol.cmd.get_model(f'resn {ligand_resn}').atom]))
                pymol.cmd.delete('*')
        return distances

    @classmethod
    def closest_hit(cls, pdb_filenames: List[str],
                    target_resi: int,
                    target_chain: str,
                    target_atomname: str,
                    ligand_resn='LIG') -> str:
        """
        This classmethod helps choose which pdb based on which is closer to a given atom.

        :param pdb_filenames:
        :param target_resi:
        :param target_chain:
        :param target_atomname:
        :param ligand_resn:
        :return:
        """
        best_d = 99999
        best_hit = -1
        for hit, d in zip(pdb_filenames,
                          cls.distance_hits(pdb_filenames, target_resi, target_chain, target_atomname, ligand_resn)):
            if d < best_d:
                best_hit = hit
                best_d = d
        return best_hit

    @classmethod
    def make_covalent(cls, smiles: str,
                      warhead_name: Optional[str] = None) -> Union[str, None]:
        """
        Convert a unreacted warhead to a reacted one in the SMILES

        :param smiles: unreacted SMILES
        :param warhead_name: name in the definitions. If unspecified it will try and guess (less preferrable)
        :return: SMILES
        """
        mol = Chem.MolFromSmiles(smiles)
        if warhead_name:
            war_defs = cls._get_warhead_definitions(warhead_name)
        else:
            war_defs = cls.warhead_definitions
        for war_def in war_defs:
            ncv = Chem.MolFromSmiles(war_def['noncovalent'])
            cv = Chem.MolFromSmiles(war_def['covalent'])
            if mol.HasSubstructMatch(ncv):
                x = Chem.ReplaceSubstructs(mol, ncv, cv, replacementConnectionPoint=0)[0]
                return Chem.MolToSmiles(x)
        else:
            return None

    # =================== extract_mols =================================================================================

    @classmethod
    def find_closest_to_ligand(cls, pdb: Chem.Mol, ligand_resn: str) -> Tuple[Chem.Atom, Chem.Atom]:
        """
        Find the closest atom to the ligand
        Warning requires the protein to be loaded as an rdkit.Chem.Mol

        :param pdb: a rdkit Chem.Mol object
        :param ligand_resn: 3 letter code
        :return: tuple of non-ligand atom and ligand atom
        """
        ligand = [atom.GetIdx() for atom in pdb.GetAtoms() if atom.GetPDBResidueInfo().GetResidueName() == ligand_resn]
        dm = Chem.Get3DDistanceMatrix(pdb)
        mini = np.take(dm, ligand, 0)
        mini[mini == 0] = np.nan
        mini[:, ligand] = np.nan
        a, b = np.where(mini == np.nanmin(mini))
        lig_atom = pdb.GetAtomWithIdx(ligand[int(a[0])])
        nonlig_atom = pdb.GetAtomWithIdx(int(b[0]))
        return (nonlig_atom, lig_atom)


    @classmethod
    def extract_mols(cls,
                     folder: str,
                     smilesdex: Dict[str, str],
                     ligand_resn: str = 'LIG',
                     regex_name: Optional[str]= None,
                     proximityBonding: bool = False,
                     throw_on_error:bool=False) -> Dict[str, Chem.Mol]:
        """
         A key requirement for Monster is a separate mol file for the parent hits.
        This is however often a pdb. This converts.
        `igor.mol_from_pose()` is similar but works on a pose. `_fix_minimized()` calls ``mol_from_pose``
        and ``copy_bonds_by_atomnames`` which does not destroy pdbinfo.
        The latter is glitchy. Use ``combine_for_bondorder``.

        See ``extract_mol`` for single.

        :param folder: folder with pdbs
        :return:
        """
        mols = {}
        for file in os.listdir(folder):
            if '.pdb' not in file:
                continue
            else:
                fullfile = os.path.join(folder, file)
                if regex_name is None:
                    name = os.path.splitext(file)[0]
                elif re.search(regex_name, file) is None:
                    continue
                else:
                    name = re.search(regex_name, file).group(1)
                if name in smilesdex:
                    smiles=smilesdex[name]
                elif throw_on_error:
                    raise ValueError(f'{name} could not be matched to a smiles.')
                else:
                    cls.journal.warning(f'{name} could not be matched to a smiles.')
                    smiles = None
                try:
                    mol = cls.extract_mol(name=name,
                                          filepath=fullfile,
                                          smiles=smiles,
                                          ligand_resn=ligand_resn,
                                          proximityBonding=proximityBonding)
                    if mol is not None:
                        mols[name] = mol
                except KeyboardInterrupt as err:
                    raise err
                except Exception as error:
                    if throw_on_error:
                        raise error
                    cls.journal.error(f'{error.__class__.__name__} for {name} - {error}')
        return mols

    @classmethod
    def extract_mol(cls,
                     name: str,
                     filepath: Optional[str] = None,
                     block: Optional[str] = None,
                     smiles: Optional[str] = None,
                     ligand_resn: str = 'LIG',
                     removeHs: bool = False,
                     proximityBonding: bool = False,
                     throw_on_error : bool = False) -> Chem.Mol:
        """
        Extracts the ligand of 3-name ``ligand_resn``
        from the PDB file ``filepath`` or from the PDB block ``block``.
        Corrects the bond order with SMILES if given.
        If there is a covalent bond with another residue the bond is kept as a ``*``/R.
        If the SMILES provided lacks the ``*`` element, the SMILES will be converted (if a warhead is matched),
        making the bond order correction okay.

        :param name: name of ligand
        :type name: str
        :param filepath: PDB file
        :type filepath: str
        :param smiles: SMILES
        :type smiles: str
        :param ligand_resn: 3letter PDB name of residue of ligand
        :type ligand_resn: str
        :param removeHs: Do you trust the hydrgens in the the PDB file?
        :type removeHs: bool
        :param throw_on_error: If an error occurs in the template step, raise error.
        :type throw_on_error: bool
        :return: rdkit Chem object
        :rtype: Chem.Mol
        """
        if filepath:
            readfun = Chem.MolFromPDBFile
            data = filepath
        elif block:
            readfun = Chem.MolFromPDBBlock
            data = block
        else:
            raise ValueError(f'Provide either a filepath or a block not neither')
        holo = readfun(data, proximityBonding=proximityBonding, removeHs=removeHs)
        if holo is None:
            cls.journal.warning(f'PDB {filepath} is problematic. Skipping sanitization.')
            holo = readfun(data, proximityBonding=False, removeHs=True, sanitize=False)
        mol = Chem.SplitMolByPDBResidues(holo, whiteList=[ligand_resn])[ligand_resn]
        mol = add_dummy_to_mol(mol, ligand_resn, holo)
        if smiles is not None:
            if '*' in Chem.MolToSmiles(mol) and '*' not in smiles:
                new_smiles = cls.make_covalent(smiles)
                if new_smiles:
                    cls.journal.info(f'{name} is covalent but ' +
                                      'a non covalent SMILES was passed, which was converted')
                    smiles = new_smiles
                else:
                    cls.journal.warning(f'{name} is covalent but ' +
                                        'a non covalent SMILES was passed, which failed to convert')
            else:
                pass
            try:
                template = Chem.MolFromSmiles(smiles)
                # template = AllChem.DeleteSubstructs(template, Chem.MolFromSmiles('*'))
                AllChem.SanitizeMol(mol)
                mol = AllChem.AssignBondOrdersFromTemplate(template, mol)
            except ValueError as error:
                if throw_on_error:
                    raise error
                else:
                    cls.journal.warning(f'{name} failed at template-guided bond order correction ' +
                                        f'- ({type(error)}: {error}).')
        mol.SetProp('_Name', name)
        return mol


    # =================== From Files ===================================================================================

    @classmethod
    def from_files(cls, folder: str) -> _VictorUtils:  # future.annotation is active, so no cyclical.
        """
        This creates an instance form the output files. Likely to be unstable.
        Assumes the checkpoints were not altered.
        And is basically for analysis only.

        :param folder: path
        :return:
        """
        cls.journal.warning('`from_files`: You really should not use this.')
        if os.path.exists(folder):
            pass # folder is fine
        elif not os.path.exists(folder) and os.path.exists(os.path.join(cls.work_path, folder)):
            folder = os.path.join(cls.work_path, folder)
        else:
            raise FileNotFoundError(f'Folder {folder} does not exist.')
        self = cls.__new__(cls)
        self.tick = float('nan')
        self.tock = float('nan')
        self.ligand_resn = ''
        self.ligand_resi = ''
        self.covalent_resn = ''
        self.covalent_resi = ''
        self.hits = []
        self.long_name = os.path.split(folder)[1]
        paramsfiles = os.path.join(folder, f'{self.long_name}.params')
        paramstemp = os.path.join(folder, f'{self.long_name}.params_template.mol')
        if os.path.exists(paramsfiles):
            self.params = Params().load(paramsfiles)
            self.unbound_pose = self.params.test()
            if os.path.exists(paramstemp):
                self.params.mol = Chem.MolFromMolFile(paramstemp, removeHs=False)
        else:
            self.params = None
        posmol = os.path.join(folder, f'{self.long_name}.positioned.mol')
        if os.path.exists(posmol):
            self.mol = Chem.MolFromMolFile(posmol, sanitize=False, removeHs=False)
        else:
            self.journal.info(f'{self.long_name} - no positioned mol')
            self.mol = None
        fragjson = os.path.join(folder, f'{self.long_name}.monster.json')
        if os.path.exists(fragjson):
            fd = json.load(open(fragjson))
            self.smiles = fd['smiles']
            self.is_covalent = True if '*' in self.smiles else False
            self.monster = Monster(self.hits)
            # self.monster.place(mol=self.mol,
            #                    attachment=None,
            #                    merging_mode='off')
            self.monster.positioned_mol = self.mol
            self.monster.positioned_mol.SetProp('_Origins', json.dumps(fd['origin']))
        else:
            self.is_covalent = None
            self.smiles = ''
            self.monster = None
            self.journal.info(f'{self.long_name} - no monster json')
            self.N_constrained_atoms = float('nan')

        #
        self.apo_pdbblock = None
        #
        self.atomnames = None
        #
        self.extra_constraint = ''
        self.pose_fx = None
        # these are calculated
        self.constraint = None
        self.unminimized_pdbblock = None
        self.igor = None
        self.minimized_pdbblock = None
        # buffers etc.
        self._warned = []
        minjson = os.path.join(folder, f'{self.long_name}.minimised.json')
        self.mrmsd = mRMSD.mock()
        if os.path.exists(minjson):
            md = json.load(open(minjson))
            self.energy_score = md["Energy"]
            self.mrmsd.mrmsd = md["mRMSD"]
            self.mrmsd.rmsds = md["RMSDs"]
            self.igor = Igor.from_pdbfile(
                pdbfile=os.path.join(folder, self.long_name + '.holo_minimised.pdb'),
                params_file=os.path.join(folder, self.long_name + '.params'),
                constraint_file=os.path.join(folder, self.long_name + '.con'))
            # victor._fix_minimized adds to igor.mol_from_pose
            self.minimized_mol = self._fix_minimized()
        else:
            self.energy_score = {'bound': {'total_score': float('nan')},
                                 'unbound': {'total_score': float('nan')}}

            self.journal.info(f'{self.long_name} - no min json')
        minmol = os.path.join(folder, f'{self.long_name}.minimised.mol')
        if os.path.exists(minmol):
            self.minimized_mol = Chem.MolFromMolFile(minmol, sanitize=False, removeHs=False)
        else:
            self.minimized_mol = None
        return self

    # =================== Guess ===================================================================================

    @classmethod
    def get_isomers(cls, mol: Chem.Mol) -> List[Chem.Mol]:
        """
        For placement operations in particular it is important to differentiate the
        isomers. Therefore requiring multiple victor calls.
        """
        enumerate_tautomers = rdMolStandardize.TautomerEnumerator().Enumerate
        enumerate_stereoisomers = EnumerateStereoisomers.EnumerateStereoisomers
        return [tauto for stereo in enumerate_stereoisomers(mol)
                for tauto in enumerate_tautomers(stereo)
                ]

    @classmethod
    def get_isomers_smiles(cls, smiles: str) -> List[str]:
        """
        Same as `get_isomers`, but with smiles.
        """
        return list(map(Chem.MolToSmiles, cls.get_isomers(Chem.MolFromSmiles(smiles))))

    @classmethod
    def guess_warhead(cls, smiles: str) -> Tuple[str, str]:
        """
        Going backwards by guessing what the warhead is.
        Normally there'd be better data handling so no guessing
        """
        if not smiles or not isinstance(smiles, str):
            return '', 'noncovalent'
        if '*' not in smiles:
            return smiles, 'noncovalent'
        mol = Chem.MolFromSmiles(smiles)
        warhead_defs = []
        for warhead_def in cls.warhead_definitions:
            if mol.HasSubstructMatch(Chem.MolFromSmiles(warhead_def['covalent'])):
                warhead_defs.append(warhead_def)
        if not warhead_defs:
            cls.journal.warning(f'Could not match {smiles} to a warhead definition in `Victor.warhead_definitions`')
            return smiles.replace('*', 'O'), 'noncovalent'
        # most complex one first!
        warhead_def = sorted(warhead_defs, key=lambda d: -len(d['covalent_atomnames']))[0]
        unrxn_mols = AllChem.ReplaceSubstructs(mol=mol,
                                               query=Chem.MolFromSmiles(warhead_def['covalent']),
                                               replacement=Chem.MolFromSmiles(warhead_def['noncovalent'])
                                               )  # noqa it is filled.
        return Chem.MolToSmiles(unrxn_mols[0]), warhead_def['name']

    def get_plip_interactions(self):
        """
        Optional, but useful to have.
        And highly experimental!
        Get the interactions from PLIP.
        """
        from .plip import SerialPLIPper
        from plip.basic import config
        # PLIP is a bit problematic with hydrogens as it doesn't like them and loses atomtypes
        config.NOHYDRO = False  # default
        clean_block = ''
        for l in self.minimized_pdbblock.split('\n'):
            if l.startswith('#'):
                break
            if '    H' in l:
                continue
            clean_block += l + '\n'
        plipper = SerialPLIPper(pdb_block=clean_block,
                          resn=self.ligand_resn,
                          resi=int(self.ligand_resi[:-1]),
                          chain=self.ligand_resi[-1])
        setattr(self, 'plipper', plipper)  # new attribute
        interaction_set = plipper.get_interaction_set(plipper.pdb_block)
        return plipper.get_interaction_counts(interaction_set)

    @staticmethod
    def to_simple_smiles(mol: Chem.Mol) -> str:
        if not isinstance(mol, Chem.Mol) or not mol.GetNumAtoms():
            return ''
        try:
            mol = AllChem.RemoveAllHs(mol)
            for atom in mol.GetAtoms():
                atom.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            return Chem.MolToSmiles(mol)
        except:
            return ''

    def migrate_sw_origins(self, row: pd.Series) -> Dict[str, Dict[int, int]]:
        """
        Given a Victor object and a SmallWorld seach result row, return the "custom_map"
        """
        # hit names to [hit to minimised]
        origins: Dict[str, Dict[int, int]] = self.monster.convert_origins_to_custom_map(forbiddance=False)
        # minimised
        query2minimized: Dict[int, int] = dict(
            enumerate(self.minimized_mol.GetSubstructMatch(Chem.MolFromSmiles(row['qrySmiles']))))
        minimized2query: Dict[int, int] = dict(zip(query2minimized.values(), query2minimized.keys()))
        # candiate
        query2candidate: Dict[int, int] = dict(enumerate(row['atomMap']))
        candidate2query: Dict[int, int] = dict(zip(query2candidate.values(), query2candidate.keys()))
        # mixing
        minimized2candidate: Dict[int, int] = {m: query2candidate.get(q, -1) for m, q in minimized2query.items()}
        origins2candidate = {name: {o: minimized2candidate.get(m, -1) for o, m in origins[name].items()} for name in
                             origins}
        return origins2candidate




"""
These are common functions moved out of Victor and Igor to reduce duplication.
"""
import logging
import warnings

from rdkit import Chem
from typing import Tuple, Union, Callable

log = logging.getLogger()

def add_dummy_to_mol(ligand: Chem.Mol,
                     ligand_resn: str,
                     holo: Chem.Mol,
                     dummy_name: str = 'CONN') -> Chem.Mol:
    """
    Given a ligand name and a holo PDB as a Chem.Mol extract the dummy atom and add it to the ligand Chem.Mol.
    Used by Victor and Igor in different points.

    :param ligand:
    :param ligand_resn:
    :param holo:
    :return:
    """
    attachment, attachee = find_attachment(holo, ligand_resn)
    if attachment is not None:  # covalent
        mod = Chem.RWMol(ligand)
        attachment.SetAtomicNum(0)  # dummy atom.
        attachment.GetPDBResidueInfo().SetName(dummy_name)
        pos = holo.GetConformer().GetAtomPosition(attachment.GetIdx())
        ni = mod.AddAtom(attachment)
        mod.GetConformer().SetAtomPosition(ni, pos)
        attachee_name = attachee.GetPDBResidueInfo().GetName()
        for atom in mod.GetAtoms():
            if atom.GetPDBResidueInfo().GetName() == attachee_name:
                ai = atom.GetIdx()
                mod.AddBond(ai, ni, Chem.BondType.SINGLE)
                break
        ligand = mod.GetMol()
        ligand.UpdatePropertyCache(strict=False)
    return ligand


def find_attachment(pdb: Chem.Mol, ligand_resn: str) -> Tuple[Union[Chem.Atom, None], Union[Chem.Atom, None]]:
    """
    Finds the two atoms in a crosslink bond without looking at LINK record

    :param pdb: a rdkit Chem object
    :param ligand_resn: 3 letter code
    :return: tuple of non-ligand atom and ligand atom
    """
    for atom in pdb.GetAtoms():
        if atom.GetPDBResidueInfo().GetResidueName() == ligand_resn:
            for neigh in atom.GetNeighbors():
                if neigh.GetPDBResidueInfo().GetResidueName() != ligand_resn:
                    attachment = neigh
                    attachee = atom
                    return attachment, attachee
    attachment = None
    attachee = None
    return attachment, attachee


def get_atomname(atom: Chem.Atom) -> str:
    return atom.GetPDBResidueInfo().GetName()

def get_equivalent(template_atom: Chem.Atom, query_mol: Chem.Mol) -> Chem.Atom:
    """
    Given an atom find the same PDB named one in the query mol.
    """
    get_by_atomname: Callable[[Chem.Mol, str], Chem.Atom] = \
        lambda mol, name: [atom for atom in mol.GetAtoms() if get_atomname(atom)[:4] == name[:4]][0]
    return get_by_atomname(query_mol, get_atomname(template_atom))

def combine_for_bondorder(template: Chem.Mol, target: Chem.Mol) -> Chem.Mol:
    target2template = {}
    for template_ai, template_atom in enumerate(template.GetAtoms()):
        target_atom: Chem.Atom = get_equivalent(template_atom, target)
        target_ai: int = target_atom.GetIdx()
        assert target_ai not in target2template.keys(), f'The atom name {get_atomname(target_atom)} is duplicated '+\
                                                   f'in the template {get_atomname(template_atom)} & '+\
                                                   f'{get_atomname(template.GetAtomWithIdx(target2template[target_ai]))}'
        target2template[target_ai] = template_ai
    new = Chem.Mol(template)
    # fix coordinates
    conf: Chem.Conformer = new.GetConformer()
    donor_conf: Chem.Conformer = target.GetConformer()
    for target_ai, template_ai in target2template.items():
        conf.SetAtomPosition(template_ai, donor_conf.GetAtomPosition(target_ai))
    # fix props
    for prop, value in target.GetPropsAsDict().items():
        if not new.HasProp(prop):
            new.SetProp(prop, value)
    for target_ai, new_ai in target2template.items():
        target_atom = target.GetAtomWithIdx(target_ai)
        new_atom = new.GetAtomWithIdx(new_ai)
        for prop, value in target_atom.GetPropsAsDict().items():
            if new_atom.HasProp(prop):
                continue
            elif isinstance(value, bool):
                new_atom.SetBoolProp(prop, value)
            elif isinstance(value, int):
                new_atom.SetIntProp(prop, value)
            elif isinstance(value, float):
                new_atom.SetDoubleProp(prop, value)
            elif isinstance(value, str):
                new_atom.SetProp(prop, value)
            else:
                pass
    return new


def copy_bonds_by_atomnames(template: Chem.Mol, target: Chem.Mol) -> bool:
    """
    Fixes bonds _inplace_ and sanity checks the mol is still the same.

    ``AllChem.AssignBondOrdersFromTemplate(template, target)``
    may not respect the atoms. Say cyclohexane -> cyclohexene alone can be mapped 12 ways by MCS,
    but only one by atomname.

    Plus ``PDBResidueInfo`` is lost by ``AllChem.AssignBondOrdersFromTemplate``
    """
    warnings.warn('This is not used anymore, but kept for reference', DeprecationWarning)
    successful = True
    for template_bond in template.GetBonds():  # type: Chem.Bond
        begin: Chem.Atom = template_bond.GetBeginAtom()
        end: Chem.Atom = template_bond.GetEndAtom()
        query_bond = target.GetBondBetweenAtoms(get_equivalent(begin, target).GetIdx(),
                                                get_equivalent(end, target).GetIdx())
        if query_bond is None:
            # This is due to the molecule being pulled apart at a ring closure.
            # A common issue with RNA minisation too.
            print(f'{get_atomname(begin)} and {get_atomname(end)} are' + \
                             ' bonded in the params but not the minimised: this is a Rosetta glitch')
            successful = False
            continue
        query_bond.SetBondType(template_bond.GetBondType())
        # heme will have crash upstream though:
        query_bond.SetBondDir(template_bond.GetBondDir())
    for template_atom in template.GetAtoms():
        target_atom: Chem.Atom = get_equivalent(template_atom, target)
        #print(get_atomname(template_atom), get_atomname(target_atom))
        target_atom.SetIsAromatic(template_atom.GetIsAromatic())
        target_atom.SetIsotope(template_atom.GetIsotope())
        target_atom.SetFormalCharge(template_atom.GetFormalCharge())
        target_atom.SetChiralTag(template_atom.GetChiralTag())
        target_atom.SetNumRadicalElectrons(template_atom.GetNumRadicalElectrons())
        target_atom.SetNumExplicitHs(template_atom.GetNumExplicitHs())
        target_atom.SetHybridization(template_atom.GetHybridization())
        target_atom.SetAtomMapNum(template_atom.GetAtomMapNum())
    # problem here:
    target.UpdatePropertyCache(strict=False)
    Chem.SanitizeMol(target)
    return successful



from .base import FragmensteinParserBase
from .monster import FragmensteinParserMonster
from .victor import FragmensteinParserVictor
from .laboratory import FragmensteinParserLaboratory
from .pipeline import FragmensteinParserPipeline
from .utils import FragmensteinParserUtils


class FragmensteinParser(FragmensteinParserMonster,
                         FragmensteinParserVictor,
                         FragmensteinParserLaboratory,
                         FragmensteinParserPipeline,
                         FragmensteinParserUtils,
                         FragmensteinParserBase):
    pass



import os

cli_default_settings = dict(
    output=os.environ.get('FRAGMENSTEIN_OUTPUT', 'output'),
    ranking=os.environ.get('FRAGMENSTEIN_RANKING', '∆∆G'),
    cutoff=float(os.environ.get('FRAGMENSTEIN_CUTOFF', 5.0)),
    quick=bool(os.environ.get('FRAGMENSTEIN_QUICK', False)),
    sw_dist=int(os.environ.get('FRAGMENSTEIN_SW_DIST', 25)),
    sw_length=int(os.environ.get('FRAGMENSTEIN_SW_LENGTH', 50)),
    sw_databases=list(os.environ.get('FRAGMENSTEIN_SW_DATABASES', 'REAL-Database-22Q1.smi.anon').split(',')),
    suffix=os.environ.get('FRAGMENSTEIN_SUFFIX', ''),
    n_cores=int(os.environ.get('FRAGMENSTEIN_N_CORES', 15)),
    combination_size=int(os.environ.get('FRAGMENSTEIN_COMBINATION_SIZE', 2)),
    top_mergers=os.environ.get('FRAGMENSTEIN_TOP_MERGERS', 500),
    max_tasks=int(os.environ.get('FRAGMENSTEIN_MAX_TASKS', 0)),
    timeout=int(os.environ.get('FRAGMENSTEIN_TIMEOUT', 240)),
    blacklist=os.environ.get('FRAGMENSTEIN_BLACKLIST', '').split(),
    workfolder=os.environ.get('FRAGMENSTEIN_WORKFOLDER', 'output'),
    linker_element=os.environ.get('FRAGMENSTEIN_LINKER_ELEMENT', 'N'),
    weights={"N_rotatable_bonds": 1,
             "\u2206\u2206G": 1,
             "interaction_uniqueness_metric": -5,
             "N_unconstrained_atoms": 0.2,
             "N_constrained_atoms": -0.05,
             "N_interactions": -1.5,
             "N_interactions_lost": 2,
             "max_hit_Tanimoto": -0.5,
             "N_PAINS": 5,
             "strain_per_HA": 1},
    supressed_exceptions=(Exception,),
)



from typing import Sequence, Union
from rdkit import Chem


class FragmensteinError(Exception):
    def fix_name(self, mol: Union[Chem.Mol, str]):
        if mol is None:
            return 'unknown'
        elif isinstance(mol, str):
            return mol
        if isinstance(mol, Chem.Mol):
            return mol.GetProp('_Name')


class ShoddyCodeError(FragmensteinError):
    """
    This is a placeholder for an error that should not happen!
    """
    pass


class FullOverlapError(FragmensteinError):
    def __init__(self,
                 message='Full overlap with no constructive outcome.',
                 hits: Sequence[Union[Chem.Mol, str]] = ()):
        self.message = message
        self.hits = hits


class PoisonError(FragmensteinError):
    def __init__(self, mol: Union[Chem.Mol, str, None] = None, indices: Sequence[int] = ()):
        self.mol = mol
        self.indices = indices

    def __str__(self):
        return f'{self.fix_name(self.mol)} has poison atoms {self.indices}'


class DistanceError(FragmensteinError):
    """
    This was formerly ConnectionError for comedy.
    """

    def __init__(self, distance: float = float('nan'), hits: Sequence[Union[Chem.Mol, str]] = (), message=None):
        self.message = message
        self.distance = distance
        self.hits = hits

    def __str__(self):
        hit_names = '+'.join([self.fix_name(h) for h in self.hits])
        distance = f'{self.distance:.2f} Å' if str(self.distance) != 'nan' else ''
        parts = [self.message, hit_names, distance]
        return ' '.join([p for p in parts if p])


class RectificationError(FragmensteinError):

    def __init__(self, message=None, mol=None):
        self.message = message
        self.mol = mol

    def __str__(self):
        return self.message



########################################################################################################################

__doc__ = \
    """
Combine = merge/join
    """

########################################################################################################################

from rdkit import Chem
from rdkit.Chem import AllChem, rdmolops

from ._collapse_ring import _MonsterRing
from ._base import _MonsterBase
from ._communal import _MonsterCommunal
from ._merge import _MonsterMerge
from molecular_rectifier import Rectifier
from ..error import DistanceError, RectificationError, FragmensteinError


########################################################################################################################

class _MonsterCombine(_MonsterRing, _MonsterMerge):

    def combine(self, keep_all: bool = True, collapse_rings: bool = True, joining_cutoff: int = 5):
        """
        Merge/links the hits. (Main entrypoint)

        :param keep_all:
        :param collapse_rings:
        :param joining_cutoff:
        :return:
        """
        # blanking
        self.unmatched = []
        self.modifications = {}
        # The following override class declared attributes.
        self.joining_cutoff = joining_cutoff
        self.throw_on_discard = keep_all
        # merge!
        if collapse_rings:
            col_hits = self.collapse_mols(self.hits)
            self.keep_copies(col_hits, 'Collapsed hits')
        else:
            col_hits = self.hits
        self.positioned_mol = self.simply_merge_hits(col_hits, linked=False)
        # remove props
        *map(self.positioned_mol.ClearProp, self.positioned_mol.GetPropNames()),  # noqa
        self.positioned_mol.SetProp('_Name', '-'.join([h.GetProp('_Name') for h in self.hits]))
        self.keep_copy(self.positioned_mol, 'merged template')
        ## Discard can happen for other reasons than disconnect
        if self.throw_on_discard and len(self.unmatched):
            raise DistanceError(hits=self.unmatched, distance=self.joining_cutoff)
        # expand and fix
        self.journal.debug(f'Merged')
        if collapse_rings:
            self.positioned_mol = self.expand_ring(self.positioned_mol)
        # bonded_as_original=False no longer needed.
        self.keep_copy(self.positioned_mol, 'expanded')
        self._join_internally(self.positioned_mol)  # will call `join_neighboring_mols` on any frags
        self.journal.debug(f'Expanded')
        try:
            self.rectify()
        except Chem.AtomValenceException:
            self.journal.info('Ring expansion while trying to appease the bonding caused an issue. Rolling back.')
            mol = Chem.RWMol(
                self.modifications['Rings expanded and original bonding restored'])  # not the novel bonding
            self._delete_collapsed(mol)
            self._detriangulate(mol)
            self.positioned_mol = self._emergency_joining(mol)
            self.rectify()
        except RecursionError:
            self.journal.critical(f'Recursion limit in rectifier')
            raise RectificationError(f'Recursion limit in rectifier', mol=self.positioned_mol)
        self.journal.debug(f'Rectified')
        return self

    def rectify(self):
        Rectifier.log = self.journal
        recto = Rectifier(self.positioned_mol)
        try:
            recto.fix()
        except FragmensteinError:
            self.journal.critical(f'This really odd cornercase: Rectifier broke the mol.')
            mol = self._emergency_joining(recto.mol)
            recto = Rectifier(mol)
            recto.fix()
        frags = AllChem.GetMolFrags(recto.mol, asMols=True)
        if len(frags) == 1:
            # all good
            self.positioned_mol = recto.mol
        else:  # the molecule is still nasty. Getting largest.
            self.positioned_mol = sorted(frags, key=lambda mol: mol.GetNumAtoms(), reverse=True)[0]
        Chem.rdmolops.AssignStereochemistryFrom3D(self.positioned_mol)
        self.keep_copies(recto.modifications, 'rectified')



from typing import (Tuple, Union, Dict)
import numpy as np
from rdkit import Chem
from rdkit import Geometry
from rdkit.Chem import rdFMCS, AllChem, rdMolAlign

from ._base import WaltonBase


class WaltonMove(WaltonBase):

    # get coordinates
    def get_point(self, atom_idx: Union[int, Geometry.Point3D], mol_idx: int = -1) -> Geometry.Point3D:
        """
        Given atom index and mol index return a point.
        If a point itself is given, return that.
        """
        if isinstance(atom_idx, Geometry.Point3D):
            return atom_idx
        elif mol_idx == -1:
            raise ValueError('If providing an atom index as a number please provide a mol_idx')
        elif isinstance(atom_idx, int):
            return self.get_mol(mol_idx).GetConformer(0).GetAtomPosition(atom_idx)
        else:
            raise TypeError(atom_idx)

    # ----------- superpose ------------------------------------------------------------------

    def superpose_by_map(self, maps: Dict[Tuple[int, int], Dict[int, int]]):
        """
        The map is a dict of a tuple of two indices: the moving one and the fixed mol.
        The items are dictionary of atom indices from the former and the correspondance in the latter.
        """
        for (moved_idx, fixed_idx), atom_map in maps.items():
            Chem.rdMolAlign.AlignMol(self.mols[moved_idx],
                                     self.mols[fixed_idx],
                                     atomMap=list(atom_map.items()))
        self.superposeed = True

    def superpose_by_mcs(self, fixed_mol_idx=-1, **mcs_settings):
        """
        Aligns by MCS. The fix molecule is the last molecule unless specified.
        """
        fixed = self.get_mol(fixed_mol_idx)
        commons = []
        for moved in self.mols[:1]:
            res: rdFMCS.MCSResult = rdFMCS.FindMCS([moved, fixed], **mcs_settings)
            common = Chem.MolFromSmarts(res.smartsString)
            Chem.SanitizeMol(common)
            mcs_map = list(zip(moved.GetSubstructMatch(common),
                               fixed.GetSubstructMatch(common)
                               )
                           )
            Chem.rdMolAlign.AlignMol(moved, fixed, atomMap=mcs_map)
            commons.append(common)
        self.superposeed = True
        return commons

    # ----- transforms ----------------------------------------

    def transform(self, mol_idx: int, transform_matrix: np.array):
        """
        Applies an affine transform 4x4 matrix to mol_idx.
        Use ``translate`` or ``rotate``.
        """
        AllChem.TransformConformer(self.get_mol(mol_idx).GetConformer(0),
                                   transform_matrix.astype(np.double)
                                   )

    def translate(self, mol_idx: int, x: float, y: float, z: float):
        """
        Move the molecule at index ``mol_idx`` of ``.mols``.
        """
        translation = np.array([[1, 0, 0, x],
                                [0, 1, 0, y],
                                [0, 0, 1, z],
                                [0, 0, 0, 1]], dtype=np.double)
        self.transform(mol_idx, translation)

    def rotate(self, mol_idx: int, theta: float, axis: str, degrees: bool = True):
        """
        Rotate a molecule around a give angle on an axis.
        This is coded terribly.

        To pivot around a given atom  has to translate the pivot point to the origin, rotate and translate back again.
        If this is gibberish, please look up affine transform 4x4 matrices.

        This is not done by Euler angle because its a one off thing.
        """
        correct_unit = lambda d: np.deg2rad(d) if degrees else d  # noqa No lambda PEP8, ha!
        theta_rad = correct_unit(theta)
        # row, column
        ori_rotation = np.zeros((4, 4)) + np.diag(np.ones(4), 0)
        if axis == 'x':  # roll
            first_row = 1
            second_row = 2
        elif axis == 'y':  # pitch
            first_row = 0
            second_row = 2
        elif axis == 'z':  # yaw
            first_row = 0
            second_row = 1
        else:
            raise ValueError(f'Axis {axis} is not x,y or z')
        # for sanity's sake
        first_col = first_row
        second_col = second_row
        ori_rotation[first_row, first_col] = np.cos(theta_rad)
        ori_rotation[first_row, second_col] = -np.sin(theta_rad)
        ori_rotation[second_row, first_col] = np.sin(theta_rad)
        ori_rotation[second_row, second_col] = np.cos(theta_rad)
        self.transform(mol_idx, ori_rotation)

    def experimental_rotate(self, mol_idx: int, angle_x: float, angle_y: float, angle_z: float, degrees: bool = True):
        """
        THIS HAS A BUG.

        Rotate a molecule around a give set of angles,
        in degrees pivoting on the origin.
        These are Tait-Bryan angles.
        There is a bug in here. Use rotate instead.

        To pivot around a given atom  has to translate the pivot point to the origin, rotate and translate back again.
        If this is gibberish, please look up affine transform 4x4 matrices.
        """

        correct_unit = lambda d: np.deg2rad(d) if degrees else d  # noqa
        alpha = correct_unit(angle_x)
        beta = correct_unit(angle_y)
        gamma = correct_unit(angle_z)
        ori_rotation = np.array([[np.cos(beta) * np.cos(gamma),
                                  np.sin(alpha) * np.sin(beta) * np.cos(gamma) - np.cos(alpha) * np.sin(gamma),
                                  np.cos(alpha) * np.sin(beta) * np.cos(gamma) + np.sin(alpha) * np.sin(gamma),
                                  0],
                                 [np.cos(beta) * np.sin(gamma),
                                  np.sin(alpha) * np.sin(beta) * np.sin(gamma) + np.cos(alpha) * np.cos(gamma),
                                  np.cos(alpha) * np.sin(beta) * np.sin(gamma) - np.sin(alpha) * np.cos(gamma),
                                  0],
                                 [-np.sin(beta),
                                  np.sin(alpha) * np.cos(beta),
                                  np.cos(alpha) * np.cos(beta),
                                  0],
                                 [0, 0, 0, 1]], dtype=np.double)
        self.transform(mol_idx, ori_rotation)



import contextlib
import operator
from typing import List
from warnings import warn

import pandas as pd
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, PandasTools
from rdkit.Chem import rdFingerprintGenerator as rdfpg
from rdkit.Chem import rdMolDescriptors as rdmd
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams, FilterCatalog, FilterCatalogEntry
from rdkit.ML.Cluster import Butina

from .._cli_defaults import cli_default_settings
from ..monster import Monster

# ----- Scoring -----------------------------------
params = FilterCatalogParams()
params.AddCatalog(FilterCatalogParams.FilterCatalogs.PAINS)
catalog = FilterCatalog(params)

def get_pains(mol) -> List[str]:
    with contextlib.suppress(Exception):
        entry: FilterCatalogEntry
        if not isinstance(mol, Chem.Mol) or mol.GetNumHeavyAtoms() == 0:
            return []
        AllChem.SanitizeMol(mol)
        return [entry.GetDescription() for entry in catalog.GetMatches(mol)]

class GetRowSimilarity:
    def __init__(self, hits):
        self.fpgen = rdfpg.GetRDKitFPGenerator()
        self.hit2fp = {h.GetProp('_Name'): self.fpgen.GetFingerprint(h) for h in hits}

    def __call__(self, row: pd.Series):
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            if not isinstance(row.minimized_mol, Chem.Mol):
                return float('nan')
            elif isinstance(row.hit_names, str):
                hit_names = row.hit_names.split(',')
            elif isinstance(row.hit_names, list):
                hit_names = row.hit_names
            else:
                return float('nan')
            fp = self.fpgen.GetFingerprint(AllChem.RemoveHs(row.minimized_mol))
            return max([DataStructs.TanimotoSimilarity(fp, self.hit2fp[name]) for name in hit_names])
        return float('nan')


class HitIntxnTallier:

    def __init__(self, hit_replacements):
        self.slim_hits = self.slim_down(hit_replacements)

    def slim_down(self, hit_replacements):
        # the bleaching was fixed cf bleached_name
        # hit_replacements['new_name'] = hit_replacements.name.str.replace('-', '_')
        # undoing bleaching
        hit_replacements['new_name'] = hit_replacements.hit_mols.apply(lambda ms: ms[0].GetProp('_Name'))
        columns = [c for c in hit_replacements.columns if isinstance(c, tuple)]
        return hit_replacements.set_index('new_name')[columns].fillna(0).copy()

    def __call__(self, row: pd.Series):
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            if not isinstance(row.minimized_mol, Chem.Mol) or isinstance(row.hit_names, float):
                return float('nan'), float('nan')
            present_tally = 0
            absent_tally = 0
            for hit_name in list(row.hit_names):
                if hit_name not in self.slim_hits.index:
                    raise Exception('Name' + hit_name)
                hit_row = self.slim_hits.loc[hit_name]
                for intxn_name, hit_value in hit_row.items():
                    if not hit_value:
                        continue
                    elif intxn_name not in row.index:
                        absent_tally += 1 if intxn_name[0] != 'hydroph_interaction' else 0.5
                    elif row[intxn_name]:
                        absent_tally += 1 if intxn_name[0] != 'hydroph_interaction' else 0.5
                    else:
                        present_tally += 1 if intxn_name[0] != 'hydroph_interaction' else 0.5
            return present_tally, absent_tally
        return float('nan'), float('nan')


class UniquenessMeter:
    def __init__(self, tallies, intxn_names, k=0.5):
        self.tallies = tallies
        self.intxn_names = intxn_names
        self.k = k

    def __call__(self, row):
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            return sum([(row[name] / self.tallies[name]) ** self.k for name in self.intxn_names if
                        row[name] and self.tallies[name]])
        return float('nan')

    def tally_interactions(self, row):
        return sum([row[c] if self.intxn_names[0] != 'hydroph_interaction' else row[c] * 0.5 for c in self.intxn_names])


class PenaltyMeter:
    def __init__(self, weights, nan_penalty=10):
        self.weights = weights
        self.nan_penalty = nan_penalty

    def __call__(self, row):
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            penalty = 0
            if row.outcome != 'acceptable':
                return float('inf')
            for col, w in self.weights.items():
                if col not in row.index:
                    warn(f'{col} column is missing from df')
                    continue
                penalty += row[col] * w if str(row[col]) != 'nan' else self.nan_penalty
            return penalty
        return float('nan')

    def make_weighted_df(self, df) -> pd.DataFrame:
        """
        Inspect whether the weights make sense
        """
        weighted = pd.DataFrame({k: df[k] * w for k, w in self.weights.items()})
        weighted['total'] = df.apply(self, axis=1)
        weighted.sort_values('total')
        return weighted


def butina_cluster(mol_list, cutoff=0.35):
    # https://github.com/PatWalters/workshop/blob/master/clustering/taylor_butina.ipynb
    fp_list = [rdmd.GetMorganFingerprintAsBitVect(AllChem.RemoveAllHs(m), 3, nBits=2048) for m in mol_list]
    dists = []
    nfps = len(fp_list)
    for i in range(1, nfps):
        sims = DataStructs.BulkTanimotoSimilarity(fp_list[i], fp_list[:i])
        dists.extend([1 - x for x in sims])
    mol_clusters = Butina.ClusterData(dists, nfps, cutoff, isDistData=True)
    cluster_id_list = [0] * nfps
    for idx, cluster in enumerate(mol_clusters, 1):
        for member in cluster:
            cluster_id_list[member] = idx
    return cluster_id_list

def UFF_Gibbs(mol):
    # free energy cost of bound conformer
    if not isinstance(mol, Chem.Mol) or mol.GetNumHeavyAtoms() == 0:
        return float('nan')
    with contextlib.suppress(cli_default_settings['supressed_exceptions']):
        AllChem.SanitizeMol(mol)
        # this is actually UFF
        copy = Chem.Mol(mol)
        return Monster.MMFF_score(None, mol, True)
    return float('nan')

class LabScore:

    @classmethod
    def score(cls,
              placements: pd.DataFrame,
              hit_replacements: pd.DataFrame,
              weights: dict,
              **settings):
        """
        This is very much a method for the CLI.
        A real Pythonic usage would be to address the individual components.
        """
        if 'minimized_mol' not in placements.columns:
            cls.Victor.journal.critical('No minimized_mol column')
            return placements
        # tanimoto
        hits: List[Chem.Mol] = hit_replacements.hit_mols.apply(operator.itemgetter(0)).to_list()
        get_similarity = GetRowSimilarity(hits)
        placements['max_hit_Tanimoto'] = placements.apply(get_similarity, axis=1)
        # properties
        m = placements.minimized_mol.apply(lambda m: m if isinstance(m, Chem.Mol) else Chem.Mol())
        # macrocyclics... yuck.
        placements['largest_ring'] = m.apply(lambda mol: max([0] + list(map(len, mol.GetRingInfo().AtomRings()))))
        # interactions
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            cls.fix_intxns(placements)
            tally_hit_intxns = HitIntxnTallier(hit_replacements)
            hit_checks = placements.apply(tally_hit_intxns, axis=1)
            placements['N_interactions_kept'] = hit_checks.apply(operator.itemgetter(0))  # .fillna(0).astype(int)
            placements['N_interactions_lost'] = hit_checks.apply(operator.itemgetter(1))  # .fillna(99).astype(int)
            intxn_names = [c for c in placements.columns if isinstance(c, tuple)]
            tallies = placements[intxn_names].sum()
            ratioed = UniquenessMeter(tallies, intxn_names, k=0.5)
            placements['interaction_uniqueness_metric'] = placements.apply(ratioed, axis=1)
            placements['N_interactions'] = placements.apply(ratioed.tally_interactions, axis=1)
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            placements['PAINSes'] = placements.minimized_mol.apply(get_pains)
            placements['N_PAINS'] = placements.PAINSes.apply(len)
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            placements['UFF_Gibbs'] = placements.minimized_mol.apply(UFF_Gibbs)
            placements['strain_per_HA'] = placements.UFF_Gibbs / (placements.N_HA + 0.0001)
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            penalize = PenaltyMeter(weights)
            placements['ad_hoc_penalty'] = placements.apply(penalize, axis=1)
        with contextlib.suppress(cli_default_settings['supressed_exceptions']):
            placements['cluster'] = butina_cluster(m.to_list())

    @staticmethod
    def export_sdf(df: pd.DataFrame,
                   penalty_col:str = 'ad_hoc_penalty',
                   filename: str = 'fragmenstein.sdf',
                   target_name: str = 'DOESNT_MATTER_UNLESS_YOU_WANT_TO_REMOVE_IT'):
        def fix(mol: Chem.Mol) -> None:
            assert isinstance(mol, Chem.Mol)
            assert mol.GetNumAtoms()
            mol.ClearComputedProps()
            for name in mol.GetPropNames():
                mol.ClearProp(name)

        with pd.option_context('mode.chained_assignment', None):
            df = df.loc[df.outcome == 'acceptable'] \
                .sort_values(penalty_col) \
                .rename(columns={c: ':'.join(map(str, c)) for c in df.columns if isinstance(c, tuple)}) \
                .reset_index() \
                .copy()
            # list of str to str w/ comma-separator
            df['ref_mols'] = df.hit_names.apply(lambda l: ','.join([v.replace(f'{target_name}-', '') for v in l]))
            df['washed_mol'] = df.minimized_mol.apply(fix)
            df['name'] = df['name'].apply(lambda v: v.split('-D68EV3CPROA')[0])
            # non str/float/ints
            not_okay = ('name', 'minimized_mol', 'ref_mols', 'washed_mol', 'mode',
                        'runtime', 'error', 'outcome',
                        'smiles',
                        'regarded', 'disregarded', 'hit_names',
                        '∆G_bound',
                        '∆G_unbound',
                        'unmin_binary',
                        'min_binary',
                        'hit_binaries',
                        'minimized_mol',
                        'hit_mols', 'unminimized_mol', 'hit_names')
        good_columns = df.columns[~df.map(lambda x: not isinstance(x, (float, str))).any()]
        extras: List[str] = [c for c in df.columns if c in good_columns and not c in not_okay]
        PandasTools.WriteSDF(df, out=filename, properties=extras, molColName='minimized_mol')


import json
from typing import List
from rdkit import Chem
from ..victor import Victor
from ..error import DistanceError
from rdkit.Chem import AllChem

class FreeVictor(Victor):
    """
    A test Victor: merge/place normally but no constraints are used.
    This has some weird behavior with Igor.
    """
    def post_params_step(self):
        for hydrogen in self.monster.positioned_mol.GetAtomsMatchingQuery(AllChem.AtomNumEqualsQueryAtom(1)):
            hydrogen.SetBoolProp('_Novel', True)

    def make_coordinate_constraints_for_placement(self, *args, **kwargs) -> str:
        return ''

    def make_coordinate_constraints_for_combination(self, *args, **kwargs) -> str:
        return ''


    def post_monster_step(self):
        """
        Damnatio memoriae!
        """
        self.monster.positioned_mol.ClearProp('_true_origins')
        self.monster.positioned_mol.ClearProp('_Origins')
        for atom in self.monster.positioned_mol.GetAtoms():
            atom.ClearProp('_ori_name')
            atom.ClearProp('_ori_i')
            atom.ClearProp('_x')
            atom.ClearProp('_y')
            atom.ClearProp('_z')


import json
from typing import List
from rdkit import Chem


class SingleVictor(Victor):
    """
    A test Victor: merge/place normally but the only use the hit number ``chosen``
    for the constraints.
    """
    chosen = 0

    def _calculate_combination_chem(self):
        """
        The rdkit part. Monster is used to combine the hits.
        """
        attachment = self._get_attachment_from_pdbblock() if self.is_covalent else None
        self._harmonize_warhead_combine()
        # TODO Does combine not need attachment??
        self.monster.modifications = self.modifications
        self.monster.combine(keep_all=self.monster_throw_on_discard,
                             collapse_rings=True,
                             joining_cutoff=self.joining_cutoff  # Å
                             )
        self.post_monster_step()  # empty overridable
        if self.monster_throw_on_discard and len(self.monster.unmatched):
            raise DistanceError(hits=self.monster.unmatched, distance=self.joining_cutoff)
        self.mol = self.monster.positioned_mol
        self.smiles = Chem.MolToSmiles(self.mol)
        # making folder.
        self.make_output_folder()

    def post_monster_step(self):
        """
        this empty overridable gets called after Monster does its thing
        in ``._calculate_placement_chem``,
        so before ``make_coordinate_constraints_for_placement``
        in ``._calculate_placement_chem`` gets called.
        The origins are not provided, but retrieved via ``monster.origin_from_mol``
        """
        keeper_name: str = self.hits[self.chosen].GetProp('_Name')
        self.true_hits = self.hits
        self.hits = [self.hits[self.chosen]]
        self.monster.hits = self.hits
        origins: List[List[str]] = self.monster.origin_from_mol()
        single_origins = [
            [parent for parent in atomic if keeper_name in parent] if isinstance(atomic, list) else []
            for atomic in origins]
        self.monster.positioned_mol.SetProp('_true_origins', json.dumps(origins))
        self.monster.positioned_mol.SetProp('_Origins', json.dumps(single_origins))
        for atom in self.monster.positioned_mol.GetAtoms():
            if atom.HasProp('_ori_name') and atom.GetProp('_ori_name') != keeper_name:
                atom.ClearProp('_ori_name')
                atom.ClearProp('_ori_i')
                atom.ClearProp('_x')
                atom.ClearProp('_y')
                atom.ClearProp('_z')


########################################################################################################################
__doc__ = \
    """
This is inherited by both place and combine via _MonsterMerge
    """

########################################################################################################################

from rdkit import Chem
from rdkit.Chem import GetPeriodicTable
from rdkit.Chem import AllChem
from rdkit.Geometry.rdGeometry import Point3D
from typing import Tuple, List, Dict, Optional, Union
import numpy as np
from .bond_provenance import BondProvenance
from ._communal import _MonsterCommunal
from ..error import DistanceError

class _MonsterJoinNeigh(_MonsterCommunal):
    def join_neighboring_mols(self, mol_A: Chem.Mol, mol_B: Chem.Mol): # noqa uppercase is fine.
        """
        Joins two molecules by first calling _find_closest to find closest.
        That method does all the thinking.
        then by calling _join_atoms.

        :param mol_A:
        :param mol_B:
        :return:Ï
        """
        # get closets atoms
        combo: Chem.RWMol
        candidates: List[Tuple[int, int, float]]
        combo, candidates = self._find_all_closest(mol_A, mol_B)  # _find_all_closest is in communal
        anchor_A, anchor_B, distance = candidates[0]
        mol = self._join_atoms(combo, anchor_A, anchor_B, distance, linking=True)
        for anchor_A, anchor_B, distance in candidates[1:]:
            mol = self._join_atoms(combo, anchor_A, anchor_B, distance, linking=False)
        mol.SetProp('_Name', mol_A.GetProp('_Name') + '~' + mol_B.GetProp('_Name'))
        return mol

    @property
    def linker_atom_zahl(self):
        """
        Getter for linker_atom_zahl
        To change set ``linker_element`` class property.
        """
        periodic_table = GetPeriodicTable()
        return periodic_table.GetAtomicNumber(self.linker_element)


    def _join_atoms(self,
                    combo: Chem.RWMol,
                    anchor_A: int,
                    anchor_B: int,
                    distance: float,
                    linking: bool=True):
        """
        extrapolate positions between. by adding linkers if needed (`linking` argument).
        """
        conf = combo.GetConformer() # noqa
        pos_A:Point3D = conf.GetAtomPosition(anchor_A)
        pos_B:Point3D = conf.GetAtomPosition(anchor_B)
        n_new:int = int(round(distance / 1.22) - 1)
        xs = np.linspace(pos_A.x, pos_B.x, n_new + 2)[1:-1]
        ys = np.linspace(pos_A.y, pos_B.y, n_new + 2)[1:-1]
        zs = np.linspace(pos_A.z, pos_B.z, n_new + 2)[1:-1]

        # ----------- correcting for ring marker atoms ---------------------------------------------------------
        def is_ring_atom(anchor: int) -> bool:
            atom = combo.GetAtomWithIdx(anchor)
            if atom.HasProp('_ori_i') and atom.GetIntProp('_ori_i') == -1:
                return True
            else:
                return False

        if is_ring_atom(anchor_A):
            distance -= 1.35 + 0.2  # Arbitrary + 0.2 to compensate for the ring not reaching (out of plane).
            n_new -= 1
            xs = xs[1:]
            ys = ys[1:]
            zs = zs[1:]

        if is_ring_atom(anchor_B):
            distance -= 1.35 + 0.2 # Arbitrary + 0.2 to compensate for the ring not reaching  (out of plane).
            n_new -= 1
            xs = xs[:-1]
            ys = ys[:-1]
            zs = zs[:-1]
        # -------------------------------------------------------------------------------------

        # notify that things could be leary.
        if distance < 0:
            self.journal.debug(f'Two ring atoms detected to be close. Joining for now.'+
                               ' They will be bonded/fused/spiro afterwards')
        # check if valid.
        if distance > self.joining_cutoff:
            msg = f'Atoms {anchor_A}+{anchor_B} are {distance} Å away. Cutoff is {self.joining_cutoff}.'
            self.journal.warning(msg)
            raise DistanceError(distance=distance)
        # place new atoms
        self.journal.debug(f'Molecules will be joined via atoms {anchor_A}+{anchor_B} ({distance} Å) '+
                           f'via the addition of {n_new} atoms.')
        previous = anchor_A
        if linking is False and n_new > 0:
            self.journal.warning(f'Was going to bond {anchor_A} and {anchor_B} but reconsidered.')
        elif previous == anchor_B:
            self.journal.warning(f'Self bonding prevented. This molecule is problematic.')
        elif linking is True and n_new <= 0:
            combo.AddBond(previous, anchor_B, Chem.BondType.SINGLE)
            new_bond = combo.GetBondBetweenAtoms(previous, anchor_B)
            BondProvenance.set_bond(new_bond, 'main_novel')
        elif linking is False and n_new <= 0:
            combo.AddBond(previous, anchor_B, Chem.BondType.SINGLE)
            new_bond = combo.GetBondBetweenAtoms(previous, anchor_B)
            BondProvenance.set_bond(new_bond, 'other_novel')
        elif linking is True and n_new > 0:
            for i in range(n_new):
                # make oxygen the first and last bridging atom.
                if i == 0 and combo.GetAtomWithIdx(anchor_A).GetSymbol() == 'C':
                    new_atomic = self.linker_atom_zahl
                elif i > 2 and i == n_new -1 and combo.GetAtomWithIdx(anchor_B).GetSymbol() == 'C':
                    new_atomic = self.linker_atom_zahl
                else:
                    new_atomic = 6
                idx = combo.AddAtom(Chem.Atom(new_atomic))
                new = combo.GetAtomWithIdx(idx)
                new.SetBoolProp('_Novel', True)
                new.SetIntProp('_ori_i', 999)
                conf.SetAtomPosition(idx, Point3D(float(xs[i]), float(ys[i]), float(zs[i])))
                combo.AddBond(idx, previous, Chem.BondType.SINGLE)
                new_bond = combo.GetBondBetweenAtoms(idx, previous)
                BondProvenance.set_bond(new_bond, 'linker')
                previous = idx
            combo.AddBond(previous, anchor_B, Chem.BondType.SINGLE)
            new_bond = combo.GetBondBetweenAtoms(previous, anchor_B)
            BondProvenance.set_bond(new_bond, 'linker')
        else:
            raise ValueError('Impossible')
        return combo.GetMol()

    def get_largest_fragment(self, mol):
        frags = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
        frags = sorted(frags, key=lambda mol: mol.GetNumAtoms(), reverse=True)
        discarded_origins = {a.GetProp('_ori_name') for m in frags[1:] for a in m.GetAtoms() if a.HasProp('_ori_name')}
        self.unmatched.extend(discarded_origins)
        return frags[0]
